"""Netcdf."""
import logging
import os
import re
from enum import Enum

import netCDF4
import numpy as np

try:
    import cfunits
except ModuleNotFoundError:
    cfunits = None
except AssertionError:
    cfunits = None
except:  # noqa
    cfunits = None


from .datetime_utils import fromtimestamp, isdatetime, utcfromtimestamp
from .geo import ConfProj, Geo
from .interpolation import Interpolation


class Netcdf(object):
    """Netcdf input."""

    def __init__(self, filename):
        """Construct NetCDF.

        Args:
            filename (str): Filename

        """
        self.filename = filename
        logging.debug("filename: %s", filename)
        self.file = netCDF4.Dataset(filename, "r")

    def nc_slice(
        self,
        var_name,
        levels=None,
        members=None,
        times=None,
        xcoords=None,
        ycoords=None,
        deaccumulate=False,
        instantanious=0.0,
        units=None,
        lev_from_ind=False,
    ):
        """Assembles a 5D field in order lon,lat,time,height,ensemble.

        Arguments:
            var_name (str): Name of field to retrieve
            levels (list): Height index. If None, return all.
            members (list): Ensemble index. If None, return all.
            times (list): Time index. If None, return all.
            xcoords: X-axis coordinates to subset
            ycoords: Y-axis coordinates to subset
            deaccumulate (bool): Deaccumulate field
            instantanious (float): Scaling factor to make an accumulated value as instantanius
            units (str): CF unit for the variable to be read
            lev_from_ind (bool): level list are indices and not values

        Raises:
            NotImplementedError: Subsetting of the input dimensions not implemented yet!
            ValueError: Times must be a list!
            ValueError: Levels must be a list!
            ValueError: Members must be a list!
            RuntimeError: No ensemble members found
            RuntimeError: cfunits not loaded!
            ValueError: Axis is not defined!")

        Returns:
            np.array: 5D array with values

        """
        var = NetCDFFileVariable(self.file, var_name)
        if xcoords is not None or ycoords is not None:
            raise NotImplementedError(
                "Subsetting of the input dimensions not implemented yet!"
            )

        tinfo = ""
        if times is not None:
            tinfo = "time " + str(times) + " in "
        logging.info("Reading %s variable %s from %s", tinfo, var.var_name, self.filename)
        times_to_read = []
        prev_time_steps = []
        if times is None:
            for i in range(0, var.times.shape[0]):
                times_to_read.append(i)
                if i > 0:
                    prev_time_steps.append(i - 1)
                else:
                    prev_time_steps.append(0)
        else:
            if not isinstance(times, (list, tuple)):
                raise ValueError("Times must be a list!")
            if isdatetime(times[0]):
                logging.debug("Time provided in call as datetime objects")
                times_in_var = var.datetimes
                for i, times_in_var_val in enumerate(times_in_var):
                    logging.debug(
                        "i %s times_in_var %s times %s", i, times_in_var_val, times
                    )
                    for tval in times:
                        # Time steps requested
                        logging.debug(
                            "i=%s, times_in_var_val=%s tval=%s", i, times_in_var_val, tval
                        )
                        if times_in_var[i] == tval:
                            times_to_read.append(i)
                            if i > 0:
                                prev_time_steps.append(i - 1)
                            else:
                                prev_time_steps.append(0)

            else:
                times_in_var = var.times
                for i in range(0, times_in_var.shape[0]):
                    for times_val in times:
                        # Time steps requested
                        if i == times_val:
                            times_to_read.append(times_val)
                            if i > 0:
                                prev_time_steps.append(i - 1)
                            else:
                                prev_time_steps.append(0)

        logging.debug("times to read %s", times_to_read)
        levels_to_read = []
        if levels is None:
            for i in range(0, var.levels.shape[0]):
                levels_to_read.append(i)
        else:
            logging.debug("Level provided in call. lev_from_ind=%s", str(lev_from_ind))
            if not isinstance(levels, (list, tuple)):
                raise ValueError("Levels must be a list!")
            levels_in_var = var.levels
            for i in range(0, levels_in_var.shape[0]):
                for level_ind in levels:
                    if lev_from_ind:
                        if i == level_ind:
                            levels_to_read.append(i)
                    else:
                        # NB! Round number to avoid round off when matching
                        if round(float(levels_in_var[i]), 5) == round(
                            float(level_ind), 5
                        ):
                            levels_to_read.append(i)

        members_to_read = []
        if members is None:
            for i in range(0, var.members.shape[0]):
                members_to_read.append(i)
        else:
            if not isinstance(members, (list, tuple)):
                raise ValueError("Members must be a list!")
            logging.debug("Ensemble members provided in call")
            members_in_var = var.members
            for i in range(0, members_in_var.shape[0]):
                for member_val in members:
                    if members_in_var[i] == member_val:
                        members_to_read.append(i)

            if len(members_to_read) == 0:
                raise RuntimeError("No ensemble members found for " + var.var_name)

        lons = var.lons
        lats = var.lats

        # Dimensions of the "problem"
        dim_x = lons.shape[0]
        dim_y = lats.shape[1]

        logging.debug("lons.shape=%s lats.shape=%s", lons.shape, lats.shape)
        geo = Geo(lons, lats)

        dim_t = max(len(times_to_read), 1)
        dim_levels = max(len(levels_to_read), 1)
        dim_members = max(len(members_to_read), 1)

        logging.debug("Dimensions in output")
        logging.debug(
            "%s %s %s %s %s",
            str(dim_x),
            str(dim_y),
            str(dim_t),
            str(dim_levels),
            str(dim_members),
        )

        lon_ind = slice(0, dim_x, 1)
        lat_ind = slice(0, dim_y, 1)
        dims = []
        prev_dims = []
        types = var.axis_types
        mapping = {}  # Map axis to output axis
        for i, type_val in enumerate(types):
            if type_val == Axis.GEOX or type_val == Axis.LON:
                dims.append(lon_ind)
                prev_dims.append(lon_ind)
                mapping[0] = i
            elif type_val == Axis.GEOY or type_val == Axis.LAT:
                dims.append(lat_ind)
                prev_dims.append(lat_ind)
                mapping[1] = i
            elif type_val == Axis.TIME:
                dims.append(times_to_read)
                prev_dims.append(prev_time_steps)
                mapping[2] = i
            elif var.is_level(type_val):
                dims.append(levels_to_read)
                prev_dims.append(levels_to_read)
                mapping[3] = i
            elif type_val == Axis.REALIZATION:
                dims.append(members_to_read)
                prev_dims.append(members_to_read)
                mapping[4] = i
            else:
                raise ValueError(str(types[i]) + " is not defined!")

        logging.debug("Read %s with dimensions: %s", var.var_name, str(dims))
        if deaccumulate:
            logging.debug("Deaccumulate previous dimensions: %s", str(prev_dims))

        logging.debug("var.varname: %s", var.var_name)
        logging.debug("dims %s", dims)
        logging.debug("self.file[var.var_name] %s", self.file[var.var_name])
        field = self.file[var.var_name][dims]
        if units is not None:
            if cfunits is None:
                raise RuntimeError("cfunits not loaded!")
            field = cfunits.Units.conform(
                field, cfunits.Units(var.units), cfunits.Units(units)
            )

        # Deaccumulation
        if deaccumulate:
            original_field = field
            previous_field = self.file[var.var_name][prev_dims]
            if units is not None:
                if cfunits is None:
                    raise RuntimeError("cfunits not loaded!")
                previous_field = cfunits.Units.conform(
                    previous_field, cfunits.Units(var.units), cfunits.Units(units)
                )
            field = np.subtract(original_field, previous_field)

        # Create instantanious values
        if instantanious > 0:
            field = np.divide(field, instantanious)

        # Add extra dimensions
        i = 0
        reverse_mapping = []
        for dim in range(0, 5):
            if dim not in mapping:
                logging.debug("Adding dimension %s", str(dim))
                field = np.expand_dims(field, len(dims) + i)
                reverse_mapping.append(len(dims) + i)
                i = i + 1
            else:
                reverse_mapping.append(mapping[dim])

        # Transpose to 5D array
        logging.debug("Transpose to 5D array")
        field = np.transpose(field, reverse_mapping)

        logging.debug("Read netcdf from %s times %s", self.filename, times)
        logging.debug("Shape of output: %s", str(field.shape))
        return field, geo

    def field(self, var_name, level=None, member=None, validtime=None, units=None):
        """Read field.

        Args:
            var_name (str): Variable name
            level (int, optional): Level. Defaults to None.
            member (int, optional): Realization. Defaults to None.
            validtime (surfex.datetime_utils.as_datetime, optional): Validtime. Defaults to None.
            units (str, optional): Units. Defaults to None.

        Returns:
            tuple: Field, Geo

        """
        if validtime is None:
            validtime = []
        else:
            validtime = [validtime]

        logging.debug("level %s member %s validtime %s", level, member, validtime)
        field, geo_in = self.nc_slice(
            var_name, levels=level, members=member, times=validtime, units=units
        )
        # Reshape to fortran 2D style
        field = np.reshape(field, [geo_in.nlons, geo_in.nlats], order="F")
        return field, geo_in

    def points(self, var, geo, validtime=None, interpolation="bilinear"):
        """Read a field and interpolate it to requested positions.

        Args:
            var (NetCDFReadVariable): NetCDF variable
            geo (surfx.geo.Geo): Geometry
            validtime (surfex.datetime_utils.as_datetime, optional): Validtime. Defaults to None.
            interpolation (str, optional): Interpolation. Defaults to "bilinear".

        Returns:
            tuple: Field, Interpolator

        """
        var_name = var.name
        level = var.level
        member = var.member
        units = var.units
        logging.debug("level %s member %s validtime %s", level, member, validtime)
        field, geo_in = self.field(
            var_name, level=level, member=member, validtime=validtime, units=units
        )
        interpolator = Interpolation(interpolation, geo_in, geo)
        field = interpolator.interpolate(field)
        return field, interpolator


class Axis(Enum):
    """Axis."""

    UNDEFINED = 0
    GEOX = 1
    GEOY = 2
    GEOZ = 3
    TIME = 4
    LON = 5
    LAT = 6
    PESSURE = 7
    HEIGHT = 8
    REFERENCETIME = 9
    REALIZATION = 10
    HYBRID = 11


class NetCDFReadVariable(object):
    """NetCDF read variable."""

    def __init__(self, name, level=None, units=None, member=None):
        """Construct object.

        Args:
            name (_type_): _description_
            level (_type_, optional): _description_. Defaults to None.
            units (_type_, optional): _description_. Defaults to None.
            member (_type_, optional): _description_. Defaults to None.

        """
        self.name = name
        self.level = level
        self.units = units
        self.member = member


class NetCDFFileVariable(object):
    """NetDF file variable."""

    def __init__(self, file_handler, var_name):
        """Construct object.

        Args:
            file_handler (_type_): _description_
            var_name (_type_): _description_

        """
        self.file = file_handler
        self.var_name = var_name

    @property
    def axis_types(self):
        """Get axis_types."""
        types = []
        if self.var_name not in self.file.variables:
            raise RuntimeError(self.var_name + " is missing in file!")

        if self.file.variables[self.var_name]:
            for dim_name in self.file.variables[self.var_name].dimensions:
                if dim_name == "longitude" or dim_name == "lon":
                    types.append(Axis.LON)
                elif dim_name == "x":
                    types.append(Axis.GEOX)
                elif dim_name == "latitude" or dim_name == "lat":
                    types.append(Axis.LAT)
                elif dim_name == "y":
                    types.append(Axis.GEOY)
                elif re.search("height[0-9]*", dim_name):
                    types.append(Axis.HEIGHT)
                elif re.search("hybrid[0-9]*", dim_name):
                    types.append(Axis.HYBRID)
                elif re.search("pressure[0-9]*", dim_name):
                    types.append(Axis.PESSURE)
                # TODO: GeoZ
                elif dim_name == "ensemble_member":
                    types.append(Axis.REALIZATION)
                elif dim_name == "time":
                    types.append(Axis.TIME)
                else:
                    types.append(Axis.UNDEFINED)
        return types

    @property
    def dim_names(self):
        """Get dim_names."""
        names = []
        if self.file.variables[self.var_name]:
            for dim_name in self.file.variables[self.var_name].dimensions:
                names.append(dim_name)
        return names

    @property
    def units(self):
        """Get units."""
        units = None
        if self.file.variables[self.var_name].units:
            units = self.file.variables[self.var_name].units
        return units

    @property
    def lats(self):
        """Get lats.

        Raises:
            RuntimeError: No latitude found

        Returns:
            np.array: 2D array of latitudes

        """
        latvals = np.array([])
        axis_types = self.axis_types
        for i, axis_type in enumerate(axis_types):
            if axis_type == Axis.LAT:
                latvals = self.file.variables[self.dim_names[i]]
                logging.warning("Assumed to 2D in (lon,lat) order")
            elif axis_type == Axis.GEOY:
                # TODO: if lat/lon are 1D, create a 2D mesh
                # TODO: Assume the name for now. Must be found in attributes
                latvals = self.file.variables["latitude"]
                latvals = np.transpose(latvals, (1, 0))

        if latvals.shape[0] == 0:
            raise RuntimeError("No latitude found for " + self.var_name)
        return latvals

    @property
    def lons(self):
        """Get lons.

        Raises:
            RuntimeError: No longitude found

        Returns:
            np.array: 2D array of longitudes

        """
        lonvals = np.array([])
        axis_types = self.axis_types
        for i, axis_type in enumerate(axis_types):
            if axis_type == Axis.LON:
                lonvals = self.file.variables[self.dim_names[i]]
            elif axis_type == Axis.GEOX:
                # TODO: if lat/lon are 1D, create a 2D mesh
                # TODO: Assume the name for now. Must be found in attributes
                lonvals = self.file.variables["longitude"]
                lonvals = np.transpose(lonvals, (1, 0))

        if lonvals.shape[0] == 0:
            raise RuntimeError("No longitude found for " + self.var_name)
        return lonvals

    @property
    def datetimes(self):
        """Get datetimes.

        Raises:
            RuntimeError: cfunits not loaded!

        Returns:
            list()

        """
        times = []
        axis_types = self.axis_types
        for i, axis_type in enumerate(axis_types):
            if axis_type == Axis.TIME:
                val = self.file.variables[self.dim_names[i]]
                for tval in val:
                    if cfunits is None:
                        raise RuntimeError("cfunits not loaded!")
                    epochtime = int(
                        cfunits.Units.conform(
                            tval,
                            cfunits.Units(val.units),
                            cfunits.Units("seconds since 1970-01-01 00:00:00"),
                        )
                    )
                    logging.debug("epoctime %s", epochtime)
                    d_t = utcfromtimestamp(epochtime)
                    logging.debug("dt %s", d_t)
                    times.append(d_t)

        if len(times) == 0:
            logging.debug("No time found for %s", self.var_name)
        return times

    @property
    def times(self):
        """Get times.

        Returns:
            np.array: 1D array of times

        """
        times = np.array([])
        axis_types = self.axis_types
        for i, axis_type in enumerate(axis_types):
            if axis_type == Axis.TIME:
                times = self.file.variables[self.dim_names[i]]

        if times.shape[0] == 0:
            logging.debug("No time found for %s", self.var_name)
        return times

    @property
    def members(self):
        """Get members.

        Returns:
            np.array: 1D array of ensemble members

        """
        members = np.array([])
        axis_types = self.axis_types
        for i, axis_type in enumerate(axis_types):
            if axis_type == Axis.REALIZATION:
                members = self.file.variables[self.dim_names[i]]

        if members.shape[0] == 0:
            logging.debug("No ensemble members found for %s", self.var_name)
        return members

    @property
    def levels(self):
        """Get levels.

        Returns:
            np.array: 1D array of levels

        """
        levels = np.array([])
        axis_types = self.axis_types
        for i, axis_type in enumerate(axis_types):
            if self.is_level(axis_type):
                levels = self.file.variables[self.dim_names[i]]

        if levels.shape[0] == 0:
            logging.debug("No levels found for %s", self.var_name)
        return levels

    @staticmethod
    def is_level(axis_type):
        """Check if is level.

        Args:
            axis_type (Axis): Acis type

        Returns:
            bool: If axis is a level type

        """
        if (
            axis_type == Axis.HEIGHT
            or axis_type == Axis.PESSURE
            or axis_type == Axis.GEOZ
            or axis_type == Axis.HYBRID
        ):
            return True
        else:
            return False


def create_netcdf_first_guess_template(
    my_variables, my_nx, my_ny, fname="raw.nc", geo=None
):
    """Create netCDF template file for first guess.

    Args:
        my_variables (_type_): _description_
        my_nx (_type_): _description_
        my_ny (_type_): _description_
        fname (str, optional): _description_. Defaults to "raw.nc".
        geo (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_

    """
    if os.path.exists(fname):
        os.remove(fname)

    my_fg = netCDF4.Dataset(fname, "w")
    my_fg.createDimension("y", my_ny)
    my_fg.createDimension("x", my_nx)
    my_fg.createDimension("time", 1)
    my_fg.createVariable("time", "f8", "time")
    my_fg.variables["time"].long_name = "time"
    my_fg.variables["time"].standard_name = "time"
    my_fg.variables["time"].units = "seconds since 1970-01-01 00:00:00 +00:00"
    my_fg.createVariable("longitude", "f8", ("y", "x"))
    my_fg.variables["longitude"].units = "degree_east"
    my_fg.variables["longitude"].long_name = "longitude"
    my_fg.variables["longitude"].standard_name = "longitude"
    my_fg.createVariable("latitude", "f8", ("y", "x"))
    my_fg.variables["latitude"].units = "degree_north"
    my_fg.variables["latitude"].long_name = "latitude"
    my_fg.variables["latitude"].standard_name = "latitude"
    my_fg.createVariable("x", "f4", "x")
    my_fg.variables["x"].long_name = "x-coordinate in Cartesian system"
    my_fg.variables["x"].standard_name = "projection_x_coordinate"
    my_fg.variables["x"].units = "m"
    my_fg.createVariable("y", "f4", "y")
    my_fg.variables["y"].long_name = "y-coordinate in Cartesian system"
    my_fg.variables["y"].standard_name = "projection_y_coordinate"
    my_fg.variables["y"].units = "m"

    standard_name = {
        "air_temperature_2m": "air_temperature",
        "relative_humidity_2m": "relative_humidity",
        "altitude": "altitude",
        "surface_snow_thickness": "surface_snow_thickness",
        "surface_soil_moisture": "surface_soil_moisture",
        "cloud_base": "cloud_base",
        "land_area_fraction": "land_area_fraction",
    }
    long_name = {
        "air_temperature_2m": "Screen level temperature (T2M)",
        "relative_humidity_2m": "Screen level relative humidity (RH2M)",
        "altitude": "Altitude",
        "surface_snow_thickness": "Surface snow thickness",
        "surface_soil_moisture": "Surface soil moisture",
        "cloud_base": "Cloud base",
        "land_area_fraction": "Land Area Fraction",
    }
    units = {
        "air_temperature_2m": "K",
        "relative_humidity_2m": "1",
        "altitude": "m",
        "surface_snow_thickness": "m",
        "surface_soil_moisture": "m3/m3",
        "cloud_base": "m",
        "land_area_fraction": "1",
    }
    fillvalue = {
        "air_temperature_2m": "9.96921e+36",
        "relative_humidity_2m": "9.96921e+36",
        "altitude": "9.96921e+36",
        "surface_snow_thickness": "9.96921e+36",
        "surface_soil_moisture": "9.96921e+36",
        "cloud_base": "9.96921e+36",
        "land_area_fraction": "9.96921e+36",
    }

    for my_var in my_variables:
        my_fg.createVariable(my_var, "f4", ("y", "x"), fill_value=fillvalue[my_var])
        my_fg.variables[my_var].long_name = long_name[my_var]
        my_fg.variables[my_var].standard_name = standard_name[my_var]
        my_fg.variables[my_var].units = units[my_var]

    # Global attributes
    if geo is not None:
        if isinstance(geo, ConfProj):
            my_fg.setncattr("gridtype", "lambert")
            my_fg.setncattr("dlon", float(geo.xdx))
            my_fg.setncattr("dlat", float(geo.xdy))
            my_fg.setncattr("projlat", float(geo.xlat0))
            my_fg.setncattr("projlat2", float(geo.xlat0))
            my_fg.setncattr("projlon", float(geo.xlon0))
            my_fg.setncattr("lonc", float(geo.xloncen))
            my_fg.setncattr("latc", float(geo.xlatcen))

    return my_fg


def read_first_guess_netcdf_file(input_file, var):
    """Read netCDF first guess file.

    Args:
        input_file (_type_): _description_
        var (_type_): _description_

    Raises:
        NotImplementedError: Only conf proj implemented when reading geo from file
        RuntimeError: cfunits not loaded!

    Returns:
        tuple:  geo, validtime, background, glafs, gelevs

    """
    file_handler = netCDF4.Dataset(input_file)
    lons = file_handler["longitude"][:]
    lats = file_handler["latitude"][:]

    if cfunits is None:
        raise RuntimeError("cfunits not loaded!")
    validtime = int(
        cfunits.Units.conform(
            file_handler["time"][:],
            cfunits.Units(file_handler["time"].units),
            cfunits.Units("seconds since 1970-01-01 00:00:00"),
        )
    )
    validtime = fromtimestamp(validtime)

    n_x = lons.shape[1]
    n_y = lons.shape[0]

    attrs = file_handler.ncattrs()
    if "gridtype" in attrs:
        if file_handler.getncattr("gridtype") == "lambert":
            from_json = {
                "nam_conf_proj_grid": {
                    "nimax": n_x,
                    "njmax": n_y,
                    "xloncen": file_handler.getncattr("lonc"),
                    "xlatcen": file_handler.getncattr("latc"),
                    "xdx": file_handler.getncattr("dlon"),
                    "xdy": file_handler.getncattr("dlat"),
                },
                "nam_conf_proj": {
                    "xlon0": file_handler.getncattr("projlon"),
                    "xlat0": file_handler.getncattr("projlat"),
                },
            }
            geo = ConfProj(from_json)
        else:
            raise NotImplementedError
    else:
        lons = np.array(np.transpose(np.reshape(lons, [n_y, n_x], order="F")))
        lats = np.array(np.transpose(np.reshape(lats, [n_y, n_x], order="F")))
        geo = Geo(lons, lats)

    background = file_handler[var][:]
    background = np.array(np.reshape(background, [n_x * n_y]))
    background = np.reshape(background, [n_y, n_x])
    background = np.transpose(background)
    fill_value = file_handler.variables[var].getncattr("_FillValue")
    logging.info("Field %s got %s. Fill with nan", var, str(fill_value))
    background[background == fill_value] = np.nan

    glafs = file_handler["land_area_fraction"][:]
    glafs = np.array(np.reshape(glafs, [n_x * n_y]))
    glafs = np.reshape(glafs, [n_y, n_x])
    glafs = np.transpose(glafs)
    gelevs = file_handler["altitude"][:]
    gelevs = np.array(np.reshape(gelevs, [n_x * n_y]))
    gelevs = np.reshape(gelevs, [n_y, n_x])
    gelevs = np.transpose(gelevs)

    file_handler.close()
    return geo, validtime, background, glafs, gelevs


def write_analysis_netcdf_file(
    filename, field, var, validtime, elevs, lafs, new_file=True, geo=None
):
    """Write analysis NetCDF file.

    Args:
        filename (_type_): _description_
        field (_type_): _description_
        var (_type_): _description_
        validtime (_type_): _description_
        elevs (_type_): _description_
        lafs (_type_): _description_
        new_file (bool, optional): _description_. Defaults to True.
        geo (_type_, optional): _description_. Defaults to None.

    Raises:
        Exception: _description_

    """
    file_handler = None
    if os.path.exists(filename):
        file_handler = netCDF4.Dataset(filename, mode="a")
    else:
        new_file = True

    if new_file:
        if geo is None:
            raise Exception("You need to provide geo to write a new file")
        file_handler = create_netcdf_first_guess_template(
            [var, "altitude", "land_area_fraction"],
            geo.nlons,
            geo.nlats,
            fname=filename,
            geo=geo,
        )
        file_handler.variables["longitude"][:] = np.transpose(geo.lons)
        file_handler.variables["latitude"][:] = np.transpose(geo.lats)
        file_handler.variables["x"][:] = list(range(0, geo.nlons))
        file_handler.variables["y"][:] = list(range(0, geo.nlats))
        file_handler.variables["altitude"][:] = np.transpose(elevs)
        file_handler.variables["land_area_fraction"][:] = np.transpose(lafs)

    file_handler["time"][:] = float(validtime.strftime("%s"))
    fill_value = file_handler.variables[var].getncattr("_FillValue")
    logging.info("Field %s got nan. Fill with %s", var, str(fill_value))
    field[np.where(np.isnan(field))] = fill_value

    file_handler[var][:] = np.transpose(field)
    file_handler.close()


def oi2soda(dtg, t2m=None, rh2m=None, s_d=None, s_m=None, output=None):
    """Convert analysis to ASCII obs file for SODA.

    Args:
        dtg (surfex.datetime_utils): Analysis time
        t2m (dict, optional): Screen level temperature var and file name. Defaults to None.
        rh2m (dict, optional): Screen level relative humidiy var and file name. Defaults to None.
        s_d (dict, optional): Snow depth var and file name. Defaults to None.
        s_m (dict, optional): Soil moisture var and file name. Defaults to None.
        output (str, optional): Output file name. Defaults to None.

    Raises:
        RuntimeError: You must specify at least one file to read from
        RuntimeError: Mismatch in ?? dimension

    """

    def check_input_to_soda_dimensions(my_nx, my_ny, nx1, ny1):

        if my_nx < 0:
            my_nx = nx1
        if my_ny < 0:
            my_ny = ny1
        if my_nx != nx1:
            raise RuntimeError(
                "Mismatch in nx dimension " + str(my_nx) + " != " + str(nx1)
            )
        if my_ny != ny1:
            raise RuntimeError(
                "Mismatch in ny dimension " + str(my_ny) + " != " + str(ny1)
            )

        return my_nx, my_ny

    cyy = dtg.strftime("%y")
    cmm = dtg.strftime("%m")
    cdd = dtg.strftime("%d")
    chh = dtg.strftime("%H")
    n_x = -1
    n_y = -1
    i = 0

    t2m_var = None
    rh2m_var = None
    sd_var = None
    sm_var = None
    if t2m is not None:
        t2m_fh = netCDF4.Dataset(t2m["file"], "r")
        logging.debug("T2m %s %s", t2m["var"], t2m_fh.variables[t2m["var"]].shape)

        i = i + 1
        n_x, n_y = check_input_to_soda_dimensions(
            n_x,
            n_y,
            t2m_fh.variables[t2m["var"]].shape[1],
            t2m_fh.variables[t2m["var"]].shape[0],
        )
        t2m_var = t2m_fh.variables[t2m["var"]][:]
        logging.debug("%s %s", t2m_var.shape, n_x * n_y)
        t2m_var = t2m_var.reshape([n_y * n_x], order="C")
        t2m_var = t2m_var.filled(fill_value=999.0)
        t2m_var = t2m_var.tolist()

    if rh2m is not None:
        rh2m_fh = netCDF4.Dataset(rh2m["file"], "r")
        logging.debug("RH2m %s %s", rh2m["var"], rh2m_fh.variables[rh2m["var"]].shape)

        i = i + 1
        n_x, n_y = check_input_to_soda_dimensions(
            n_x,
            n_y,
            rh2m_fh.variables[rh2m["var"]].shape[1],
            rh2m_fh.variables[rh2m["var"]].shape[0],
        )
        rh2m_var = rh2m_fh.variables[rh2m["var"]][:]
        rh2m_var = rh2m_var.reshape([n_y * n_x], order="C")
        rh2m_var = rh2m_var.filled(fill_value=999.0)
        rh2m_var = rh2m_var.tolist()

    if s_d is not None:
        sd_fh = netCDF4.Dataset(s_d["file"], "r")
        logging.debug("SD %s %s", s_d["var"], sd_fh.variables[s_d["var"]].shape)

        i = i + 1
        n_x, n_y = check_input_to_soda_dimensions(
            n_x,
            n_y,
            sd_fh.variables[s_d["var"]].shape[1],
            sd_fh.variables[s_d["var"]].shape[0],
        )

        sd_var = sd_fh.variables[s_d["var"]][:]
        sd_var = sd_var.reshape([n_y * n_x], order="C")
        sd_var = sd_var.filled(fill_value=999.0)
        sd_var = sd_var.tolist()

    if s_m is not None:
        sm_fh = netCDF4.Dataset(s_m["file"], "r")
        logging.debug("SM %s %s", s_m["var"], sm_fh.variables[s_m["var"]].shape)

        i = i + 1
        n_x, n_y = check_input_to_soda_dimensions(
            n_x,
            n_y,
            sm_fh.variables[s_m["var"]].shape[1],
            sm_fh.variables[s_m["var"]].shape[0],
        )

        sm_var = sm_fh.variables[s_m["var"]][:]
        sm_var = sm_var.reshape([n_y * n_x], order="C")
        sm_var = sm_var.filled(fill_value=999.0)
        sm_var = sm_var.tolist()

    if i == 0:
        raise RuntimeError("You must specify at least one file to read from!")

    if output is None:
        output = (
            "OBSERVATIONS_" + str(cyy) + str(cmm) + str(cdd) + "H" + str(chh) + ".DAT"
        )

    with open(output, mode="w", encoding="utf-8") as out:
        for i in range(0, n_x * n_y):
            line = ""
            if t2m_var is not None:
                line = line + " " + str(t2m_var[i])
            if rh2m_var is not None:
                line = line + " " + str(rh2m_var[i])
            if sd_var is not None:
                line = line + " " + str(sd_var[i])
            if sm_var is not None:
                line = line + " " + str(sm_var[i])
            line = line + "\n"
            out.write(line)
            logging.debug("i %s", i)


def read_cryoclim_nc(infiles):
    """Read crycoclim netCDF file.

    Args:
        infiles (list): Input files.

    Raises:
        RuntimeError: "No files were read properly"

    Returns:
        tuple: grid_lons, grid_lats, grid_snow_class

    """
    grid_lons = None
    grid_lats = None
    grid_snow_class = None
    for filename in infiles:
        if os.path.exists(filename):
            logging.info("Reading: %s", filename)
            ncf = netCDF4.Dataset(filename, "r")
            grid_lons = ncf["lon"][:]
            grid_lats = ncf["lat"][:]
            grid_snow_class_read = ncf["classed_product"][:]
            if grid_snow_class is None:
                grid_snow_class = grid_snow_class_read
            grid_snow_class[grid_snow_class_read == 1] = 1
            grid_snow_class[grid_snow_class_read == 0] = 0
            ncf.close()
        else:
            logging.warning("Warning file %s does not exists", filename)

    if grid_lons is None or grid_lats is None or grid_snow_class is None:
        raise RuntimeError("No files were read properly")

    return grid_lons, grid_lats, grid_snow_class


def read_sentinel_nc(infiles):
    """Read sentinel nc files.

    Args:
        infiles (list): Input files.

    Raises:
        RuntimeError: "No files were read properly"

    Returns:
        tuple: longitudes, latitudes, soil moisture
    """
    grid_lons = None
    grid_lats = None
    grid_sm = None
    for filename in infiles:
        if os.path.exists(filename):
            logging.info("Reading: %s", filename)
            nch = netCDF4.Dataset(filename, "r")
            grid_lons = nch["LON"][:]
            grid_lats = nch["LAT"][:]
            grid_sm = nch["surface_soil_moisture"][:]
            nch.close()
        else:
            logging.warning("Warning file %s does not exists", filename)

    if grid_lons is None or grid_lats is None or grid_sm is None:
        raise RuntimeError("No files were read properly")

    return grid_lons, grid_lats, grid_sm



"""bufr treatment."""
import logging
import sys
from math import exp

import numpy as np

try:
    import eccodes  # type: ignore
except ImportError:
    eccodes = None
    logging.warning("ECCODES not found. Needed for bufr reading")
except RuntimeError:
    eccodes = None
    logging.warning("ECCODES not found. Needed for bufr reading")
# Needed in Python 3.5
except Exception:
    logging.warning("Could not load eccodes")
    eccodes = None


from .datetime_utils import as_datetime_args
from .obs import ObservationSet
from .observation import Observation


class BufrObservationSet(ObservationSet):
    """Create observation data set from bufr observations."""

    def __init__(
        self,
        bufrfile,
        variables,
        valid_dtg,
        valid_range,
        lonrange=None,
        latrange=None,
        label="bufr",
        use_first=False,
    ):
        """Initialize a bufr observation set.

        Args:
            bufrfile (str): Full path of the bufr file to read
            variables(list): Variables to read
            valid_dtg (datetime.datetime): DTG string with valid time
            valid_range (datetime.timedelta): The allowed time range
            lonrange (list): Allowed range of longitudes [min, max]
            latrange (list): Allowed range of latitides [min, max]
            label (str): A label for the resulting observations set
            use_first (bool): Use only the first valid observation for a point if more are found

        Raises:
            RuntimeError: ECCODES not found. Needed for bufr reading
            NotImplementedError: Not implemented

        """
        if lonrange is None:
            lonrange = [-180, 180]
        if latrange is None:
            latrange = [-90, 90]

        if eccodes is None:
            raise RuntimeError("ECCODES not found. Needed for bufr reading")
        logging.debug(eccodes.__file__)

        # open bufr file
        file_handler = open(bufrfile, mode="rb")

        # define the keys to be printed
        keys = [
            "latitude",
            "localLatitude",
            "longitude",
            "localLongitude",
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "heightOfStationGroundAboveMeanSeaLevel",
            "heightOfStation",
            "stationNumber",
            "blockNumber",
        ]
        nerror = {}
        ntime = {}
        nundef = {}
        ndomain = {}
        nobs = {}
        for var in variables:
            if var == "relativeHumidityAt2M":
                keys.append("airTemperatureAt2M")
                keys.append("dewpointTemperatureAt2M")
                keys.append(
                    "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=2"
                    "/airTemperature"
                )
                keys.append(
                    "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=1.5"
                    "/airTemperature"
                )
                keys.append(
                    "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=2"
                    "/dewpointTemperature"
                )
                keys.append(
                    "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=1.5"
                    "/dewpointTemperature"
                )
            elif var == "airTemperatureAt2M":
                keys.append("airTemperatureAt2M")
                keys.append(
                    "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=2"
                    "/airTemperature"
                )
                keys.append(
                    "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=1.5"
                    "/airTemperature"
                )
            else:
                keys.append(var)
            nerror.update({var: 0})
            ntime.update({var: 0})
            nundef.update({var: 0})
            ndomain.update({var: 0})
            nobs.update({var: 0})

        logging.info("Reading %s", bufrfile)
        logging.info("Looking for keys: %s", str(keys))
        cnt = 0
        observations = list()

        # loop for the messages in the file
        not_decoded = 0
        registry = {}
        while 1:
            # get handle for message
            bufr = eccodes.codes_bufr_new_from_file(file_handler)
            if bufr is None:
                break

            # we need to instruct ecCodes to expand all the descriptors
            # i.e. unpack the data values
            try:
                eccodes.codes_set(bufr, "unpack", 1)
                decoded = True
            except eccodes.CodesInternalError as err:
                not_decoded = not_decoded + 1
                logging.error('Error with key="unpack" : %s', err.msg)
                decoded = False

            # print the values for the selected keys from the message
            if decoded:
                lat = np.nan
                local_lat = np.nan
                lon = np.nan
                local_lon = np.nan
                value = np.nan
                elev = np.nan
                year = -1
                month = -1
                day = -1
                hour = -1
                minute = -1
                stid = "NA"
                station_number = -1
                block_number = -1
                t2m = np.nan
                td2m = np.nan
                s_d = np.nan
                temp = np.nan
                t_d = np.nan
                c_b = np.nan
                for key in keys:
                    try:
                        logging.debug("Decode: %s", key)
                        val = eccodes.codes_get(bufr, key)
                        logging.debug("Got: %s=%s", key, val)
                        if (
                            val == eccodes.CODES_MISSING_DOUBLE
                            or val == eccodes.CODES_MISSING_LONG
                        ):
                            val = np.nan
                        if key == "latitude":
                            lat = val
                            if lat < -90 or lat > 90:
                                lat = np.nan
                        if key == "longitude":
                            lon = val
                            if lon < -180 or lon > 180:
                                lon = np.nan
                        if key == "localLatitude":
                            local_lat = val
                            if local_lat < -90 or local_lat > 90:
                                local_lat = np.nan
                        if key == "localLongitude":
                            local_lon = val
                            if local_lon < -180 or local_lon > 180:
                                local_lon = np.nan
                        if key == "year":
                            year = val
                        if key == "month":
                            month = val
                        if key == "day":
                            day = val
                        if key == "hour":
                            hour = val
                        if key == "minute":
                            minute = val
                        if key == "heightOfStation":
                            if not np.isnan(val):
                                elev = val
                        if key == "heightOfStationGroundAboveMeanSeaLevel":
                            if not np.isnan(val):
                                elev = val
                        if key == "stationNumber":
                            station_number = val
                        if key == "blockNumber":
                            block_number = val
                        if key == "airTemperatureAt2M":
                            t2m = val
                        if (
                            key
                            == "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=2"
                            "/airTemperature"
                            or key
                            == "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=1.5"
                            "/airTemperature"
                        ):
                            temp = val
                        if key == "dewpointTemperatureAt2M":
                            td2m = val
                        if (
                            key
                            == "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=2"
                            "/dewpointTemperature"
                            or key
                            == "/heightOfSensorAboveLocalGroundOrDeckOfMarinePlatform=1.5"
                            "/dewpointTemperature"
                        ):
                            t_d = val
                        if key == "totalSnowDepth":
                            s_d = val
                        if key == "heightOfBaseOfCloud":
                            c_b = val

                    except eccodes.CodesInternalError:
                        logging.debug('Report does not contain key="%s"', key)

                got_pos = True
                if np.isnan(lat):
                    if np.isnan(local_lat):
                        got_pos = False
                    else:
                        lat = local_lat
                if np.isnan(lon):
                    if np.isnan(local_lon):
                        got_pos = False
                    else:
                        lon = local_lon

                if got_pos:
                    # Assign value to var
                    pos = f"{lon:.5f}:{lat:.5f}"
                    for var in variables:
                        exists = False
                        if use_first:
                            if pos in registry:
                                if var in registry[pos]:
                                    exists = registry[pos][var]
                            else:
                                registry.update({pos: {}})
                        if not exists:
                            logging.debug("Pos does not exist %s %s", pos, var)
                            if var == "relativeHumidityAt2M":
                                if not np.isnan(t2m) and not np.isnan(td2m):
                                    try:
                                        value = self.td2rh(td2m, t2m)
                                        value = value * 0.01
                                    except Exception:
                                        logging.debug("Got exception for %s:", var)
                                        value = np.nan
                                else:
                                    if not np.isnan(temp) and not np.isnan(t_d):
                                        try:
                                            value = self.td2rh(t_d, temp)
                                            value = value * 0.01
                                        except Exception:
                                            logging.debug(
                                                "Got exception for %s",
                                                var,
                                            )
                                            value = np.nan
                            elif var == "airTemperatureAt2M":
                                if np.isnan(t2m):
                                    if not np.isnan(temp):
                                        value = temp
                                else:
                                    value = t2m
                            elif var == "totalSnowDepth":
                                value = s_d
                            elif var == "heightOfBaseOfCloud":
                                value = c_b
                            else:
                                raise NotImplementedError(
                                    f"Var {var} is not coded! Please do it!"
                                )
                        else:
                            logging.debug("Pos already exists %s %s", pos, var)

                        all_found = True
                        if np.isnan(lat):
                            all_found = False
                        if np.isnan(lon):
                            all_found = False
                        if year == -1:
                            all_found = False
                        if month == -1:
                            all_found = False
                        if day == -1:
                            all_found = False
                        if hour == -1:
                            all_found = False
                        if minute == -1:
                            all_found = False
                        if np.isnan(elev):
                            all_found = False
                        if np.isnan(value):
                            all_found = False
                        if not all_found:
                            nerror.update({var: nerror[var] + 1})

                        logging.debug(
                            "Check on position in space and time %s %s %s %s %s %s",
                            lon,
                            lonrange[0],
                            lonrange[1],
                            lat,
                            latrange[0],
                            latrange[1],
                        )
                        if (
                            latrange[0] <= lat <= latrange[1]
                            and lonrange[0] <= lon <= lonrange[1]
                        ):
                            obs_dtg = as_datetime_args(
                                year=year, month=month, day=day, hour=hour, minute=minute
                            )
                            if not np.isnan(value):
                                if self.inside_window(obs_dtg, valid_dtg, valid_range):
                                    logging.debug(
                                        "Valid DTG for station %s %s %s %s %s %s %s %s",
                                        obs_dtg,
                                        valid_dtg,
                                        valid_range,
                                        lon,
                                        lat,
                                        value,
                                        elev,
                                        stid,
                                    )
                                    if station_number > 0 and block_number > 0:
                                        stid = str((block_number * 1000) + station_number)
                                    observations.append(
                                        Observation(
                                            obs_dtg,
                                            lon,
                                            lat,
                                            value,
                                            elev=elev,
                                            stid=stid,
                                            varname=var,
                                        )
                                    )
                                    if use_first:
                                        registry[pos].update({var: True})
                                    nobs.update({var: nobs[var] + 1})
                                else:
                                    ntime.update({var: ntime[var] + 1})
                            else:
                                nundef.update({var: nundef[var] + 1})
                        else:
                            ndomain.update({var: ndomain[var] + 1})

                cnt += 1

                if (cnt % 1000) == 0:
                    logging.info(".", end="")
                    sys.stdout.flush()

            # delete handle
            eccodes.codes_release(bufr)

        logging.info("\nFound %s/%s", str(len(observations)), str(cnt))
        logging.info("Not decoded: %s", str(not_decoded))
        for var in variables:
            logging.info("\nObservations for var=%s: %s", var, str(nobs[var]))
            logging.info(
                "Observations removed because of domain check: %s", str(ndomain[var])
            )
            logging.info(
                "Observations removed because of not being defined/found: %s",
                str(nundef[var]),
            )
            logging.info(
                "Observations removed because of time window: %s", str(ntime[var])
            )
            logging.info(
                "Messages not containing information on all keys: %s", str(nerror[var])
            )
        # close the file
        file_handler.close()

        ObservationSet.__init__(self, observations, label=label)

    @staticmethod
    def td2rh(t_d, temp, kelvin=True):
        """Convert dew point to temperature.

        Args:
            t_d (float): Dew point temperature
            temp (float): Temperature
            kelvin (bool, optional): Kelvin. Defaults to True.

        Raises:
            RuntimeError: Dew point temperature is probably not Kelvin
            RuntimeError: Temperature is probably not Kelvin

        Returns:
            float: Relative humidity (percent)

        """
        if kelvin:
            if t_d < 100:
                raise RuntimeError("Dew point temperature is probably not Kelvin")
            if temp < 100:
                raise RuntimeError("Temperature is probably not Kelvin")
            t_d = t_d - 273.15
            temp = temp - 273.15

        r_h = 100 * (
            exp((17.625 * t_d) / (243.04 + t_d)) / exp((17.625 * temp) / (243.04 + temp))
        )
        if r_h > 110 or r_h < 1:
            logging.warning(
                "\nWARNING: Calculated rh to %s from %s and %s. Set it to missing",
                str(r_h),
                str(t_d),
                str(temp),
            )
            r_h = np.nan
        elif r_h > 100:
            logging.warning(
                "\nWARNING: Calculated rh to %s from %s and %s.",
                str(r_h),
                str(t_d),
                str(temp) + " Truncate to 100%",
            )
            r_h = 100
        return r_h

    @staticmethod
    def inside_window(obs_dtg, valid_dtg, valid_range):
        """Check if inside window.

        Args:
            obs_dtg (as_datetime): Observation datetime
            valid_dtg (as_datetime): Valid datetime
            valid_range (as_timedelta): Window

        Returns:
            bool: True if inside window

        """
        if valid_dtg is None:
            return True
        else:
            if (valid_dtg - valid_range) <= obs_dtg <= (valid_dtg + valid_range):
                return True
            else:
                return False



"""Command line interfaces."""
import os
import sys
from argparse import Action, ArgumentParser

import yaml

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None


from . import __version__


class LoadFromFile(Action):
    """Load arguments from a file."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Override __call__ method."""
        with values as f_h:
            # parse arguments in the file and store them in the target namespace
            parser.parse_args(f_h.read().split(), namespace)


def parse_args_create_forcing(argv):
    """Parse arguments to create forcing.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Create offline forcing")
    parser.add_argument("dtg_start", type=str, help="Start DTG", nargs="?")
    parser.add_argument("dtg_stop", type=str, help="Stop DTG", nargs="?")
    parser.add_argument(
        "-d",
        dest="domain",
        type=str,
        help="Domain file describing the points or locations",
        nargs="?",
        required=False,
        default=None,
    )
    parser.add_argument(
        "--harmonie",
        action="store_true",
        default=False,
        help="Surfex configuration (domain) created from Harmonie environment",
    )
    parser.add_argument(
        "--config_exp_surfex",
        dest="config_exp_surfex",
        type=str,
        help="Toml configuration file for surfex settings potentially "
        + "used if --harmonie is set",
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "-fb", type=str, help="First base time unless equal to dtg_start", default=None
    )
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="user_config",
        type=str,
        help="Configuration file "
        + "in yaml format describing customized variable setup",
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "-t", "--timestep", type=int, help="Surfex time step", default=3600, nargs="?"
    )
    parser.add_argument(
        "-ci",
        "--cache_interval",
        type=int,
        help="clear cached fields after..",
        default=3600,
        nargs="?",
    )
    parser.add_argument(
        "-i",
        "--input_format",
        type=str,
        help="Default input file format",
        default="netcdf",
        choices=["netcdf", "grib1", "grib2", "surfex", "fa"],
    )
    parser.add_argument(
        "-ig",
        "--input_geo",
        dest="geo_input",
        type=str,
        help="Default input geometry if needed",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-o",
        "--output_format",
        type=str,
        help="Output file format",
        default="nc4",
        choices=["netcdf", "nc4", "ascii"],
        nargs="?",
    )
    parser.add_argument("-a", dest="analysis", action="store_true", default=False)
    parser.add_argument(
        "--interpolation",
        dest="interpolation",
        required=False,
        default="bilinear",
        choices=["nearest", "bilinear"],
    )
    parser.add_argument("-of", type=str, help="Output file name", default=None, nargs="?")
    parser.add_argument(
        "-p", "--pattern", type=str, help="Filepattern", default=None, nargs="?"
    )
    parser.add_argument(
        "--zref",
        type=str,
        help="Temperature/humidity reference height",
        default="ml",
        choices=["ml", "screen"],
    )
    parser.add_argument(
        "--uref",
        type=str,
        help="Wind reference height: screen/ml/",
        default="ml",
        choices=["ml", "screen"],
    )
    parser.add_argument("--debug", help="Show debug information", action="store_true")
    parser.add_argument(
        "--single", help="Print single time step twice", action="store_true"
    )
    parser.add_argument("--version", action="version", version=__version__)

    group_ta = parser.add_argument_group("TA", description="Air temperature [K]")
    group_ta.add_argument(
        "--ta",
        type=str,
        help="Input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex"],
    )
    group_ta.add_argument(
        "--ta_converter",
        type=str,
        help="Converter function to air temperature",
        default="none",
        choices=["none"],
    )

    group_qa = parser.add_argument_group("QA", description="Specific humidity")
    group_qa.add_argument(
        "--qa",
        type=str,
        help="Input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex"],
    )
    group_qa.add_argument(
        "--qa_converter",
        type=str,
        help="Converter function to specific humidity",
        default="none",
        choices=["none", "rh2q", "rh2q_mslp"],
    )

    group_ps = parser.add_argument_group("PS", description="Surface air pressure [Pa]")
    group_ps.add_argument(
        "--ps",
        type=str,
        help="Surface air pressure input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_ps.add_argument(
        "--ps_converter",
        type=str,
        help="Converter function to surface air pressure",
        default="none",
        choices=["none", "mslp2ps"],
    )

    group_dir_sw = parser.add_argument_group(
        "DIR_SW", description="Direct shortwave radiation"
    )
    group_dir_sw.add_argument(
        "--dir_sw",
        type=str,
        help="Direct short wave radiation input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_dir_sw.add_argument(
        "--dir_sw_converter",
        type=str,
        help="Converter function to direct short wave radiation",
        default="none",
        choices=["none", "analysis"],
    )

    group_sca_sw = parser.add_argument_group(
        "SCA_SW", description="Scattered short wave radiation flux"
    )
    group_sca_sw.add_argument(
        "--sca_sw",
        type=str,
        help="Scattered short wave radiation input format",
        default="default",
        choices=["netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_sca_sw.add_argument(
        "--sca_sw_converter",
        type=str,
        help="Converter function to scattered shortwave radiation flux",
        default="none",
        choices=["none"],
    )

    group_lw = parser.add_argument_group("LW", description="Long wave radiation flux")
    group_lw.add_argument(
        "--lw",
        type=str,
        help="Long wave radiation input format",
        default="default",
        choices=["netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_lw.add_argument(
        "--lw_converter",
        type=str,
        help="Converter function to long wave radiation flux",
        default="none",
        choices=["none", "analysis"],
    )

    group_rain = parser.add_argument_group("RAIN", description="Rainfall rate")
    group_rain.add_argument(
        "--rain",
        type=str,
        help="Input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex"],
    )
    group_rain.add_argument(
        "--rain_converter",
        type=str,
        help="Converter function to rainfall rate",
        default="totalprec",
        choices=["none", "totalprec", "calcrain"],
    )

    group_snow = parser.add_argument_group("SNOW", description="Snowfall rate")
    group_snow.add_argument(
        "--snow",
        type=str,
        help="Input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex"],
    )
    group_snow.add_argument(
        "--snow_converter",
        type=str,
        help="Converter function to snowfall rate",
        default="none",
        choices=["none", "calcsnow", "snowplusgraupel"],
    )

    group_wind = parser.add_argument_group("WIND", description="Wind speed")
    group_wind.add_argument(
        "--wind",
        type=str,
        help="Input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex"],
    )
    group_wind.add_argument(
        "--wind_converter",
        type=str,
        help="Converter function to windspeed",
        default="windspeed",
        choices=["none", "windspeed"],
    )

    group_wind_dir = parser.add_argument_group("WIND_DIR", description="Wind direction")
    group_wind_dir.add_argument(
        "--wind_dir",
        type=str,
        help="Input format",
        default="default",
        choices=["default", "netcdf", "grib1", "grib2", "surfex"],
    )
    group_wind_dir.add_argument(
        "--wind_dir_converter",
        type=str,
        help="Converter function to wind direction",
        default="winddir",
        choices=["none", "winddir"],
    )

    group_co2 = parser.add_argument_group("CO2", description="Carbon dioxide")
    group_co2.add_argument(
        "--co2",
        type=str,
        help="CO2 input format",
        default="default",
        choices=["netcdf", "grib1", "constant", "grib2", "surfex"],
    )
    group_co2.add_argument(
        "--co2_converter",
        type=str,
        help="Converter function to carbon dioxide",
        default="none",
        choices=["none"],
    )

    group_zs = parser.add_argument_group("ZS", description="Surface geopotential")
    group_zs.add_argument(
        "--zsoro",
        type=str,
        help="ZS input format",
        default="default",
        choices=["netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_zs.add_argument(
        "--zsoro_converter",
        type=str,
        help="Converter function to ZS",
        default="none",
        choices=["none", "phi2m"],
    )

    group_zval = parser.add_argument_group(
        "ZREF", description="Reference height for temperature " "and humidity"
    )
    group_zval.add_argument(
        "--zval",
        type=str,
        help="ZREF input format",
        default="default",
        choices=["netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_zval.add_argument(
        "--zval_converter",
        type=str,
        help="Converter function to ZREF",
        default="none",
        choices=["none"],
    )

    group_uval = parser.add_argument_group(
        "UREF", description="Reference height for wind"
    )
    group_uval.add_argument(
        "--uval",
        type=str,
        help="UREF input format",
        default="default",
        choices=["netcdf", "grib1", "grib2", "surfex", "constant"],
    )
    group_uval.add_argument(
        "--uval_converter",
        type=str,
        help="Converter function to UREF",
        default="none",
        choices=["none"],
    )

    if len(argv) < 4:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})

    user_config = {}
    if "user_config" in kwargs and kwargs["user_config"] is not None:
        user_config = (
            yaml.safe_load(open(kwargs["user_config"], mode="r", encoding="utf-8")) or {}
        )
    kwargs.update({"user_config": user_config})

    # Find name of global config file
    root = __file__
    if os.path.islink(root):
        root = os.path.realpath(root)
    base = os.path.dirname(os.path.abspath(root))
    yaml_config = base + "/cfg/config.yml"

    default_conf = yaml.safe_load(
        open(yaml_config, mode="r", encoding="utf-8")
    ) or sys.exit(1)
    kwargs.update({"config": default_conf})
    return kwargs


def parse_args_modify_forcing(argv):
    """Parse arguments to modify forcing.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Modify offline forcing NetCDF file")
    parser.add_argument(
        "-i",
        "--input_file",
        type=str,
        help="Input forcing file",
        nargs="?",
        required=True,
    )
    parser.add_argument(
        "-t",
        "--time_step",
        type=str,
        help="Time step ",
        nargs="?",
        required=False,
        default=-1,
    )
    parser.add_argument(
        "-o",
        "--output_file",
        type=str,
        help="Output forcing file",
        nargs="?",
        required=True,
    )
    parser.add_argument("variables", type=str, nargs="+", help="Variables to substitute")

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_qc2obsmon(argv):
    """Parse arguments for qc2obsmon.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Create SQLite data base for obsmon")
    parser.add_argument("dtg", type=str, help="YYYYMMDDHH")
    parser.add_argument("varname", type=str, help="Variable name")
    parser.add_argument("qc", type=str, help="QC dataset JSONfile")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "--operator",
        type=str,
        help="Obs operator",
        choices=["bilinear", "nearest"],
        default="bilinear",
        required=False,
    )
    parser.add_argument("--fg_file", type=str, help="First guess file", required=True)
    parser.add_argument("--an_file", type=str, help="Analysis file", required=True)
    parser.add_argument("--file_var", type=str, help="File variable", required=True)
    parser.add_argument(
        "-o", dest="output", type=str, nargs="?", help="output file", default="ecma.db"
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_first_guess_for_oi(argv):
    """Parse arguments for firstguess4oi.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Create first guess file for gridpp")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-dtg", dest="dtg", type=str, help="Date (YYYYMMDDHH)", required=True
    )
    parser.add_argument(
        "-i", "--inputfile", type=str, default=None, help="Default input file", nargs="?"
    )
    parser.add_argument(
        "-if", dest="inputformat", type=str, help="Input file format", default="grib2"
    )
    parser.add_argument(
        "-d", dest="domain", type=str, help="Domain", required=False, default=None
    )
    parser.add_argument(
        "--harmonie",
        action="store_true",
        default=False,
        help="Surfex configuration (domain) created from Harmonie environment",
    )

    parser.add_argument(
        "-t2m_file", type=str, default=None, help="File with T2M", nargs="?"
    )
    parser.add_argument(
        "-t2m_format",
        type=str,
        default=None,
        help="File format for file with T2M",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "-t2m_converter",
        type=str,
        default="none",
        help="Converter for T2M",
        nargs="?",
        choices=["none", "tap"],
    )
    parser.add_argument(
        "-rh2m_file", type=str, default=None, help="File with RH2M", nargs="?"
    )
    parser.add_argument(
        "-rh2m_format",
        type=str,
        default=None,
        help="File format for file with RH2M",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "-rh2m_converter",
        type=str,
        default="none",
        help="Converter for RH2M",
        nargs="?",
        choices=["none", "rhp"],
    )

    parser.add_argument(
        "-sd_file", type=str, default=None, help="Snow depth file", nargs="?"
    )
    parser.add_argument(
        "-sd_format",
        type=str,
        default=None,
        help="Snow depth file format",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "--sd_converter",
        type=str,
        default="none",
        help="",
        nargs="?",
        choices=["none", "sweclim", "swe2sd", "sdp"],
    )

    parser.add_argument(
        "-cb_file", type=str, default=None, help="Cloud base file", nargs="?"
    )
    parser.add_argument(
        "-cb_format",
        type=str,
        default=None,
        help="Cloud base file format",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "--cb_converter",
        type=str,
        default="cloud_base",
        help="",
        nargs="?",
        choices=["cloud_base"],
    )

    parser.add_argument(
        "-sm_file", type=str, default=None, help="Soil moisture file", nargs="?"
    )
    parser.add_argument(
        "-sm_format",
        type=str,
        default=None,
        help="Soil moisture file format",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "--sm_converter",
        type=str,
        default="none",
        help="",
        nargs="?",
        choices=["none", "smp"],
    )

    parser.add_argument(
        "-laf_file",
        type=str,
        default=None,
        help="Land area fraction grib file",
        nargs="?",
    )
    parser.add_argument(
        "-laf_format",
        type=str,
        default=None,
        help="Snow depth file format",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "--laf_converter",
        type=str,
        default="nature_town",
        help="",
        nargs="?",
        choices=["none", "sea2land", "nature_town"],
    )

    parser.add_argument(
        "-altitude_file", type=str, default=None, help="SURFEX grib file", nargs="?"
    )
    parser.add_argument(
        "-altitude_format",
        type=str,
        default=None,
        help="Snow depth file format",
        nargs="?",
        choices=["grib1", "grib2", "netcdf", "surfex", "fa"],
    )
    parser.add_argument(
        "--altitude_converter",
        type=str,
        default="phi2m",
        help="",
        nargs="?",
        choices=["none", "phi2m"],
    )

    parser.add_argument(
        "-o", dest="output", type=str, help="Output file", default="raw.nc"
    )
    parser.add_argument(
        "--config",
        "-c",
        dest="input_config",
        type=str,
        help="YAML config file",
        default="first_guess.yml",
        nargs="?",
    )
    parser.add_argument(
        "variables",
        nargs="+",
        choices=[
            "air_temperature_2m",
            "relative_humidity_2m",
            "surface_snow_thickness",
            "cloud_base",
            "surface_soil_moisture",
        ],
        help="Variables to create first guess for",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_masterodb(argv):
    """Parse the command line input arguments for masterodb.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="SURFEX for MASTERRODB")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument("--version", action="version", version=f"surfex {__version__}")
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument(
        "--wrapper", "-w", type=str, default="", help="Execution wrapper command"
    )
    parser.add_argument(
        "--harmonie",
        action="store_true",
        default=False,
        help="Surfex configuration created from Harmonie environment",
    )
    parser.add_argument(
        "--pgd", type=str, nargs="?", required=True, help="Name of the PGD file"
    )
    parser.add_argument(
        "--prep", type=str, nargs="?", required=True, help="Name of the PREP file"
    )
    parser.add_argument(
        "--force", "-f", action="store_true", default=False, help="Force re-creation"
    )
    parser.add_argument("--rte", "-r", required=True, nargs="?")
    parser.add_argument("--config", "-c", required=False, nargs="?")
    parser.add_argument(
        "--system_file_paths",
        "-s",
        required=True,
        nargs="?",
        help="Input file paths on your system",
    )
    parser.add_argument("--namelist_path", "-n", required=True, nargs="?")
    parser.add_argument(
        "--domain", type=str, required=False, help="JSON file with domain"
    )
    parser.add_argument("--dtg", type=str, required=False, default=None)
    parser.add_argument("--output", "-o", type=str, required=False, default=None)
    parser.add_argument(
        "--only_archive", action="store_true", default=False, help="Only call archiving"
    )
    parser.add_argument(
        "--tolerate_missing",
        action="store_true",
        default=False,
        help="Tolerate missing files",
    )
    parser.add_argument(
        "--print_namelist",
        action="store_true",
        default=False,
        help="Print namelsist used",
    )
    parser.add_argument(
        "--mode", "-m", type=str, required=True, choices=["forecast", "canari"]
    )
    parser.add_argument(
        "--archive",
        "-a",
        required=False,
        default=None,
        nargs="?",
        help="JSON file with archive output",
    )
    parser.add_argument(
        "--binary",
        "-b",
        required=False,
        default=None,
        nargs="?",
        help="Full path of MASTERODB binary",
    )

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_surfex_binary(argv, mode):
    """Parse the command line input arguments for surfex binary.

    Args:
        argv (list): List with arguments.
        mode(str): Type of surfex binary

    Raises:
        NotImplementedError: Mode not implemented

    Returns:
        dict: Parsed arguments.

    """
    pert = False
    need_pgd = True
    need_prep = True
    if mode == "pgd":
        need_pgd = False
        need_prep = False
        desc = "Create physiography for SURFEX (PGD)"
    elif mode == "prep":
        need_prep = False
        desc = "Prepare initial conditions for SURFEX"
    elif mode == "offline":
        desc = "Run Offline SURFEX"
    elif mode == "soda":
        desc = "Run SURFEX data assimilation (SODA)"
    elif mode == "perturbed":
        pert = True
        desc = "Run perturbed Offline SURFEX"
    else:
        raise NotImplementedError(mode + " is not implemented!")

    parser = ArgumentParser(description=desc)
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument(
        "--wrapper", "-w", type=str, default="", help="Execution wrapper command"
    )
    if need_pgd:
        parser.add_argument(
            "--pgd", type=str, nargs="?", required=True, help="Name of the PGD file"
        )
    if need_prep:
        parser.add_argument(
            "--prep", type=str, nargs="?", required=True, help="Name of the PREP file"
        )
    if mode == "prep":
        parser.add_argument("--prep_file", required=False, default=None, nargs="?")
        parser.add_argument("--prep_filetype", required=False, default=None, nargs="?")
        parser.add_argument("--prep_pgdfile", required=False, default=None, nargs="?")
        parser.add_argument("--prep_pgdfiletype", required=False, default=None, nargs="?")
    if mode == "offline" or mode == "perturbed":
        parser.add_argument(
            "--forc_zs",
            action="store_true",
            default=False,
            help="Set model ZS to forcing ZS",
        )
        parser.add_argument("--forcing_dir", required=False, default=None, nargs="?")
    parser.add_argument("--force", "-f", action="store_true", help="Force re-creation")
    parser.add_argument(
        "--harmonie",
        action="store_true",
        default=False,
        help="Surfex configuration created from Harmonie environment",
    )
    parser.add_argument(
        "--print_namelist", action="store_true", default=False, help="Print namelist used"
    )
    parser.add_argument(
        "--tolerate_missing",
        action="store_true",
        default=False,
        help="Tolerate missing files",
    )
    parser.add_argument(
        "--masterodb",
        action="store_true",
        default=False,
        help="Input file written by masterodb",
    )
    parser.add_argument("--rte", "-r", required=True, nargs="?")
    parser.add_argument("--config", "-c", required=False, nargs="?")
    parser.add_argument(
        "--system_file_paths",
        "-s",
        required=True,
        nargs="?",
        help="Input file paths on your system",
    )
    parser.add_argument("--namelist_path", "-n", required=True, nargs="?")
    parser.add_argument(
        "--domain", type=str, required=False, help="JSON file with domain"
    )
    parser.add_argument("--output", "-o", type=str, required=True)
    parser.add_argument("--dtg", type=str, required=False, default=None)
    if pert:
        parser.add_argument("--pert", "-p", type=int, required=False, default=None)
        parser.add_argument(
            "--negpert", action="store_true", default=False, help="Negative perturbation"
        )
    parser.add_argument(
        "--archive",
        "-a",
        type=str,
        required=False,
        default=None,
        nargs="?",
        help="JSON file with archive output",
    )
    parser.add_argument("binary", type=str, help="Command to run")

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_create_namelist(argv):
    """Parse the command line input arguments for creating a namelist.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Create namelist")
    parser.add_argument("--version", action="version", version=__version__)
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument(
        "--wrapper", "-w", type=str, default="", help="Execution wrapper command"
    )
    parser.add_argument("mode", type=str, help="Type of namelist")
    parser.add_argument("--method", required=False, default="blocks", nargs="?")
    parser.add_argument("--prep_file", required=False, default=None, nargs="?")
    parser.add_argument("--prep_filetype", required=False, default=None, nargs="?")
    parser.add_argument("--prep_pgdfile", required=False, default=None, nargs="?")
    parser.add_argument("--prep_pgdfiletype", required=False, default=None, nargs="?")
    parser.add_argument(
        "--forc_zs", action="store_true", default=False, help="Set model ZS to forcing ZS"
    )
    parser.add_argument("--forcing_dir", required=False, default=None, nargs="?")
    parser.add_argument(
        "--harmonie",
        action="store_true",
        default=False,
        help="Surfex configuration created from Harmonie environment",
    )
    parser.add_argument(
        "--system_file_paths",
        "-s",
        required=True,
        nargs="?",
        help="Input file paths on your system",
    )
    parser.add_argument("--config", "-c", required=False, nargs="?")
    parser.add_argument("--namelist_path", "-n", required=True, nargs="?")
    parser.add_argument(
        "--domain", type=str, required=False, help="JSON file with domain"
    )
    parser.add_argument("--output", "-o", type=str, required=False)
    parser.add_argument("--dtg", type=str, required=False, default=None)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_gridpp(argv):
    """Parse the command line input arguments for gridpp.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Create horisontal OI analysis")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-i",
        "--input_file",
        type=str,
        help="Input NetCDF file with all variables",
        required=True,
    )
    parser.add_argument(
        "-obs",
        "--obs_file",
        type=str,
        help="Input JSON file with QC observations",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output_file",
        type=str,
        help="Output NetCDF file with all variables",
        required=True,
    )
    parser.add_argument("-v", "--var", type=str, help="Variable", required=True)
    parser.add_argument("-hor", dest="hlength", type=float, required=True)
    parser.add_argument(
        "-vert", dest="vlength", type=float, default=100000, required=False
    )
    parser.add_argument(
        "--wlength", dest="wlength", type=float, default=0.0, required=False
    )
    parser.add_argument(
        "--maxLocations", dest="max_locations", type=int, default=20, required=False
    )
    parser.add_argument(
        "--elevGradient",
        dest="elev_gradient",
        type=float,
        default=0,
        required=False,
        choices=[0, -0.0065],
    )
    parser.add_argument(
        "--epsilon", dest="epsilon", type=float, default=0.25, required=False
    )
    parser.add_argument(
        "--minvalue", dest="minvalue", type=float, default=None, required=False
    )
    parser.add_argument(
        "--maxvalue", dest="maxvalue", type=float, default=None, required=False
    )
    parser.add_argument(
        "--only_diff",
        action="store_true",
        help="Only write differences to file",
        required=False,
        default=False,
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_titan(argv):
    """Parse the command line input arguments for titan.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Do quality control of observations")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-i",
        "--input_file",
        type=str,
        help="Input json file with observation sets and test settings",
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output_file",
        type=str,
        help="Output json file with quality checked observations",
        required=False,
        default="qc_obs.json",
    )
    parser.add_argument(
        "-v", "--variable", type=str, required=True, help="Observation variable"
    )
    parser.add_argument("--indent", type=int, default=None, help="Indent")
    parser.add_argument(
        "-dtg", type=str, help="Date time group YYYYMMDDHH", required=True
    )
    parser.add_argument(
        "--harmonie",
        action="store_true",
        default=False,
        help="Surfex configuration created from Harmonie environment",
    )
    parser.add_argument(
        "tests", nargs="+", type=str, help="Which tests to run and order to run"
    )
    parser.add_argument(
        "--blacklist",
        dest="blacklist_file",
        type=str,
        required=False,
        default=None,
        help="JSON file with blacklist",
    )
    parser.add_argument(
        "--domain", type=str, required=False, default=None, help="JSON file with domain"
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_oi2soda(argv):
    """Parse the command line input arguments for oi2soda.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Create ASCII input for SODA from gridpp files")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "--t2m_file", type=str, help="NetCDF file for T2M", required=False, default=None
    )
    parser.add_argument(
        "--t2m_var",
        type=str,
        help="NetCDF variable name for T2M",
        required=False,
        default="air_temperature_2m",
    )
    parser.add_argument(
        "--rh2m_file", type=str, help="NetCDF file for RH2M", required=False, default=None
    )
    parser.add_argument(
        "--rh2m_var",
        type=str,
        help="NetCDF variable name for RH2M",
        required=False,
        default="relative_humidity_2m",
    )
    parser.add_argument(
        "--sd_file", type=str, help="NetCDF file for SD", required=False, default=None
    )
    parser.add_argument(
        "--sd_var",
        type=str,
        help="NetCDF variable name for SD",
        required=False,
        default="surface_snow_thickness",
    )
    parser.add_argument(
        "--sm_file", type=str, help="NetCDF file for SM", required=False, default=None
    )
    parser.add_argument(
        "--sm_var",
        type=str,
        help="NetCDF variable name for SM",
        required=False,
        default="surface_soil_moisture",
    )
    parser.add_argument("dtg", nargs="?", type=str, help="DTG", default=None)
    parser.add_argument("-o", dest="output", type=str, help="Output file", default=None)
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) < 3:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_lsm_file_assim(argv):
    """Parse the command line input arguments for land-sea-mask for assimilation.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser(description="Create ASCII LSM input for SODA")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument("--file", type=str, help="Input file name", required=True)
    parser.add_argument("--fileformat", type=str, help="Input fileformat", required=True)
    parser.add_argument(
        "--var",
        type=str,
        help="Variable in input file",
        required=False,
        default="air_temperature_2m",
    )
    parser.add_argument(
        "--converter",
        type=str,
        help="Converter for variable",
        required=False,
        default="none",
    )
    parser.add_argument("--dtg", type=str, help="DTG", default=None, required=False)
    parser.add_argument("--domain", type=str, help="Domain", required=True)
    parser.add_argument("-o", dest="output", type=str, help="Output file", default=None)
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) < 3:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})

    return kwargs


def parse_args_hm2pysurfex(argv):
    """Parse the command line input arguments for hm2pysurfex.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("hm2pysurfex")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-c", dest="config", type=str, required=True, help="PySurfex config file"
    )
    parser.add_argument(
        "-e",
        dest="environment",
        type=str,
        required=False,
        default=None,
        help="Environment if not taken from running environment",
    )
    parser.add_argument(
        "-o",
        dest="output",
        type=str,
        required=False,
        default=None,
        help="Output toml file",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_bufr2json(argv):
    """Parse the command line input arguments for bufr2json.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("bufr2json")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument("-b", dest="bufr", type=str, required=True, help="Bufr file")
    parser.add_argument(
        "-v", dest="vars", nargs="+", type=str, required=True, help="Variables"
    )
    parser.add_argument(
        "-o", dest="output", type=str, required=True, help="Output JSON file"
    )
    parser.add_argument(
        "-dtg", dest="dtg", type=str, required=True, help="DTG (YYYYMMDHH)"
    )
    parser.add_argument(
        "--indent", dest="indent", type=int, required=False, default=None, help="Indent"
    )
    parser.add_argument(
        "-range",
        dest="valid_range",
        type=int,
        help="Valid range in seconds",
        default=3600,
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_obs2json(argv):
    """Parse the command line input arguments for obs2json.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("obs2json")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-t",
        dest="obs_type",
        type=str,
        required=True,
        help="Observations type",
        choices=["bufr", "netatmo", "frost", "obsoul", "json"],
    )
    parser.add_argument(
        "-i", dest="inputfile", type=str, nargs="+", required=True, help="inputfile(s)"
    )
    parser.add_argument(
        "-v", dest="vars", nargs="+", type=str, required=True, help="Variables"
    )
    parser.add_argument(
        "-o", dest="output", type=str, required=True, help="Output JSON file"
    )
    parser.add_argument(
        "-dtg", dest="obs_time", type=str, required=True, help="DTG (YYYYMMDHH)"
    )
    parser.add_argument(
        "--indent", dest="indent", type=int, required=False, default=None, help="Indent"
    )
    parser.add_argument(
        "--pos_t_range",
        dest="pos_t_range",
        type=int,
        help="Valid range in seconds after obs_time",
        default=3600,
    )
    parser.add_argument(
        "--neg_t_range",
        dest="neg_t_range",
        type=int,
        help="Valid range in seconds before obs_time",
        default=3600,
    )
    parser.add_argument(
        "--label", dest="label", type=str, required=False, default=None, help="Label"
    )
    parser.add_argument(
        "--unit", dest="unit", type=str, required=False, default=None, help="Unit (FROST)"
    )
    parser.add_argument(
        "--level",
        dest="level",
        type=str,
        required=False,
        default=None,
        help="Level (FROST)",
    )
    parser.add_argument(
        "--obtypes",
        dest="obtypes",
        type=str,
        required=False,
        default=None,
        help="Obtypes (obsoul)",
    )
    parser.add_argument(
        "--subtypes",
        dest="subtypes",
        type=str,
        required=False,
        default=None,
        help="Subtypes (obsoul)",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_plot_points(argv):
    """Parse the command line input arguments for plotting points.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Plot points")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-g",
        "--geo",
        dest="geo",
        type=str,
        help="Domain/points json geometry definition file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-v",
        "--variable",
        dest="variable",
        type=str,
        help="Variable name",
        required=False,
    )
    parser.add_argument(
        "-i",
        "--inputfile",
        dest="inputfile",
        type=str,
        help="Input file",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-it",
        "--inputtype",
        dest="inputtype",
        type=str,
        help="Filetype",
        default="surfex",
        required=False,
        choices=["netcdf", "grib1", "grib2", "surfex", "obs"],
    )
    parser.add_argument(
        "-t",
        "--validtime",
        dest="validtime",
        type=str,
        help="Valid time",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output file",
        default=None,
        required=False,
    )
    parser.add_argument("--no-contour", dest="no_contour", action="store_true")
    parser.add_argument(
        "--interpolator", type=str, default="nearest", required=False, help="Interpolator"
    )
    grib = parser.add_argument_group("grib", "Grib1/2 settings (-it grib1 or -it grib2)")
    grib.add_argument(
        "--indicatorOfParameter",
        type=int,
        help="Indicator of parameter [grib1]",
        default=None,
    )
    grib.add_argument(
        "--timeRangeIndicator", type=int, help="Time range indicator [grib1]", default=0
    )
    grib.add_argument(
        "--levelType", type=str, help="Level type [grib1/grib2]", default="sfc"
    )
    grib.add_argument("--level", type=int, help="Level [grib1/grib2]", default=0)
    grib.add_argument("--discipline", type=int, help="Discipline [grib2]", default=None)
    grib.add_argument(
        "--parameterCategory", type=int, help="Parameter category [grib2]", default=None
    )
    grib.add_argument(
        "--parameterNumber", type=int, help="ParameterNumber [grib2]", default=None
    )
    grib.add_argument(
        "--typeOfStatisticalProcessing",
        type=int,
        help="TypeOfStatisticalProcessing [grib2]",
        default=-1,
    )

    sfx = parser.add_argument_group("Surfex", "Surfex settings (-it surfex)")
    sfx.add_argument(
        "--sfx_type",
        type=str,
        help="Surfex file type",
        default=None,
        choices=[None, "forcing", "ascii", "nc", "netcdf", "texte"],
    )

    sfx.add_argument("--sfx_patches", type=int, help="Patches [ascii/texte]", default=-1)
    sfx.add_argument("--sfx_layers", type=int, help="Layers [ascii/texte]", default=-1)
    sfx.add_argument(
        "--sfx_datatype",
        type=str,
        help="Datatype [ascii]",
        choices=["string", "float", "integer"],
        default="float",
    )
    sfx.add_argument("--sfx_interval", type=str, help="Interval [texte]", default=None)
    sfx.add_argument("--sfx_basetime", type=str, help="Basetime [texte]", default=None)
    sfx.add_argument(
        "--sfx_geo_input",
        type=str,
        default=None,
        help="JSON file with domain defintion [forcing/netcdf/texte]",
    )

    obs = parser.add_argument_group("Observations", "Observation settings (scatter plot)")
    obs.add_argument(
        "--obs_type",
        type=str,
        help="Observation source type (-it obs)",
        choices=[None, "json", "bufr", "frost", "netatmo"],
        default=None,
    )
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_set_geo_from_obs_set(argv):
    """Parse the command line input arguments for setting geo from obs set.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Set a point geometry from an observation set")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-v", type=str, dest="variable", help="Variable name", required=True
    )
    parser.add_argument(
        "-t", dest="validtime", help="Validtime (YYYYMMDDHH)", required=True
    )
    parser.add_argument(
        "-i", type=str, dest="inputfile", help="Input file", required=False
    )
    parser.add_argument(
        "-it",
        type=str,
        dest="obs_type",
        help="Input type",
        required=True,
        choices=["netatmo", "frost", "bufr", "json"],
    )
    parser.add_argument(
        "--lonrange",
        type=str,
        dest="lonrange",
        help="Longitude range",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--latrange",
        type=str,
        dest="latrange",
        help="Latitude range",
        default=None,
        required=False,
    )
    parser.add_argument("-o", type=str, dest="output", help="Output file", required=True)
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_set_geo_from_stationlist(argv):
    """Parse the command line input arguments for setting geo from station list.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Set a point geometry from a stationlist")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument("stationlist", type=str, help="Station list")
    parser.add_argument(
        "--lonrange",
        type=str,
        dest="lonrange",
        help="Longitude range",
        default=None,
        required=False,
    )
    parser.add_argument(
        "--latrange",
        type=str,
        dest="latrange",
        help="Latitude range",
        default=None,
        required=False,
    )
    parser.add_argument("-o", type=str, dest="output", help="Output file", required=True)
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument("--version", action="version", version=__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_merge_qc_data(argv):
    """Parse the command line input arguments for merge of qc data.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser()
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-i",
        type=str,
        nargs="+",
        dest="filenames",
        help="Input QC JSON files",
        required=True,
    )
    parser.add_argument(
        "-t", dest="validtime", help="Validtime (YYYYMMDDHH)", required=True
    )
    parser.add_argument("--indent", type=int, help="Indent in output", default=None)
    parser.add_argument("-o", type=str, dest="output", help="Output file", required=True)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_timeseries2json(argv):
    """Parse the command line input arguments for time series to json.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Convert a time series to json")
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-v", "--varname", dest="varname", type=str, help="Variable name", required=True
    )
    parser.add_argument(
        "-lons",
        dest="lons",
        type=float,
        nargs="+",
        help="Longitudes",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-lats",
        dest="lats",
        type=float,
        nargs="+",
        help="Latitudes",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-stids",
        dest="stations",
        type=str,
        nargs="+",
        help="Longitudes",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-stations",
        dest="stationlist",
        type=str,
        help="Longitudes",
        default=None,
        required=False,
    )
    parser.add_argument(
        "-i",
        "--filepattern",
        dest="filepattern",
        type=str,
        help="Input file",
        default="",
        required=False,
    )
    parser.add_argument(
        "-it",
        "--inputtype",
        dest="inputtype",
        type=str,
        help="Input type (format)",
        default="surfex",
        required=False,
        choices=["netcdf", "grib1", "grib2", "surfex", "obs"],
    )
    parser.add_argument(
        "-start", dest="start", type=str, help="Start time (YYYYMMDDHH)", required=True
    )
    parser.add_argument(
        "-end", dest="end", type=str, help="End time (YYYYMMDDHH)", required=True
    )
    parser.add_argument(
        "-int",
        dest="interval",
        type=int,
        help="Interval in seconds",
        required=False,
        default=3600,
    )
    parser.add_argument(
        "-indent", dest="indent", type=int, help="Indent", required=False, default=None
    )
    parser.add_argument(
        "-fcint",
        dest="fcint",
        type=int,
        help="Interval between analysis in seconds",
        required=False,
        default=3 * 3600,
    )
    parser.add_argument(
        "-file_inc",
        dest="file_inc",
        type=int,
        help="Interval between analysis in seconds",
        required=False,
        default=3 * 3600,
    )
    parser.add_argument(
        "-offset",
        dest="offset",
        type=int,
        help="Offset into next forecast by seconds",
        required=False,
        default=0,
    )
    parser.add_argument(
        "-sfx",
        dest="sfx_type",
        type=str,
        help="Input type for surfex files",
        default=None,
        required=False,
        choices=[None, "forcing", "ascii", "nc", "netcdf", "texte"],
    )
    parser.add_argument(
        "-geo",
        dest="geo_in",
        type=str,
        help="JSON file with geometry needed for some surfex file types",
        required=False,
        default=None,
    )
    parser.add_argument(
        "-obs",
        dest="obs_set",
        type=str,
        help="Input type",
        default=None,
        required=False,
        choices=[None, "json", "bufr", "frost", "netatmo", "titan"],
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output image",
        default=None,
        required=False,
    )

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_cryoclim_pseudoobs(argv):
    """Parse the command line input arguments for cryoclim pseudo obs.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Create CRYOCLIM pseudo-obs")
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-v",
        "--varname",
        dest="varname",
        type=str,
        help="Variable name",
        default="surface_snow_thickness",
        required=False,
    )
    parser.add_argument(
        "-fg",
        dest="fg_file",
        type=str,
        help="First guess file",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-i",
        dest="infiles",
        type=str,
        nargs="+",
        help="Infiles",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-step",
        dest="thinning",
        type=int,
        help="Thinning step",
        required=False,
        default=4,
    )
    parser.add_argument(
        "-indent", dest="indent", type=int, help="Indent", required=False, default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output observation set",
        default=None,
        required=False,
    )

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_sentinel_obs(argv):
    """Parse the command line input arguments for sentinel observations.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Create Sentinel-1 obs")
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-v",
        "--varname",
        dest="varname",
        type=str,
        help="Variable name",
        default="surface_soil_moisture",
        required=False,
    )
    parser.add_argument(
        "-fg",
        dest="fg_file",
        type=str,
        help="First guess file",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-i",
        dest="infiles",
        type=str,
        nargs="+",
        help="Infiles",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-step",
        dest="thinning",
        type=int,
        help="Thinning step",
        required=False,
        default=4,
    )
    parser.add_argument(
        "-indent", dest="indent", type=int, help="Indent", required=False, default=None
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output image",
        default=None,
        required=False,
    )

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_args_shape2ign(argv):
    """Parse the command line input arguments for shape fiel to ign.

    Args:
        argv (list): List with arguments.

    Returns:
        dict: Parsed arguments.

    """
    parser = ArgumentParser("Convert NVE shape files to IGN geometry")
    parser.add_argument(
        "--debug", action="store_true", help="Debug", required=False, default=False
    )
    parser.add_argument(
        "--options", type=open, action=LoadFromFile, help="Load options from file"
    )
    parser.add_argument(
        "-c",
        "--catchment",
        dest="catchment",
        type=str,
        help="Catchment name",
        default="None",
        required=False,
    )
    parser.add_argument(
        "-i",
        dest="infile",
        type=str,
        help="Infile/directory",
        default=None,
        required=True,
    )
    parser.add_argument(
        "-r",
        dest="ref_proj",
        type=str,
        help="Reference projection (domain file)",
        default=None,
        required=True,
    )
    parser.add_argument(
        "--indent", dest="indent", type=str, help="Indent", default=None, required=False
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        help="Output json geometry file",
        default=None,
        required=False,
    )

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def parse_set_domain(argv):
    """Parse the command line input arguments."""
    parser = ArgumentParser()

    parser.add_argument("--version", action="version", version=f"surfex {__version__}")
    parser.add_argument("--domain", "-d", required=True, type=str, help="Name of domain")
    parser.add_argument("--domains", required=True, type=str, help="Domain definitions")
    parser.add_argument(
        "--harmonie", action="store_true", help="Domain in harmonie definition"
    )
    parser.add_argument(
        "--indent", required=False, default=2, type=int, help="Indented output"
    )
    parser.add_argument("--output", "-o", required=True, nargs="?")
    parser.add_argument("--debug", help="Show debug information", action="store_true")

    if len(argv) == 1:
        parser.print_help()
        sys.exit()

    return parser.parse_args(argv)



"""obs."""
import json
import logging
import os

import numpy as np
import requests

try:
    import cfunits
except ModuleNotFoundError:
    cfunits = None
except AssertionError:
    cfunits = None
except:  # noqa
    cfunits = None


from .datetime_utils import as_datetime, as_datetime_args, as_timedelta, utcfromtimestamp
from .interpolation import gridpos2points, inside_grid
from .observation import Observation
from .titan import QCDataSet, dataset_from_file


class ObservationSet(object):
    """Set of observations."""

    def __init__(self, observations, label=""):
        """Create an observation set.

        Args:
            observations (list): Observation objects.
            label (str, optional): Name of set. Defaults to "".

        """
        self.observations = observations
        self.size = len(self.observations)
        self.label = label
        self.index_pos = {}
        self.index_stid = {}

    def get_stid_index(self, stid):
        """Get station ID index.

        Args:
            stid (str): Station ID.

        Returns:
            int: Found index

        """
        stid = str(stid)
        if stid in self.index_stid:
            return self.index_stid[stid]
        else:
            return None

    def get_pos_index(self, lon, lat):
        """Get position index.

        Args:
            lon (float):Longitude
            lat (float): Latitude

        Returns:
            int: Found position index.
        """
        lon = Observation.format_lon(lon)
        lat = Observation.format_lat(lat)
        pos = lon + ":" + lat
        if pos in self.index_pos:
            return self.index_pos[pos]
        else:
            return None

    def get_obs(self):
        """Get observations.

        Returns:
            (list, list , list, list, list, list, list): times, lons, lats, stids, elevs,
                                                         values, varnames
        """
        obs2vectors = np.vectorize(Observation.obs2vectors)
        logging.debug("Obs dim %s", len(self.observations))
        if len(self.observations) > 0:
            times, lons, lats, stids, elevs, values, varnames = obs2vectors(
                self.observations
            )

            for point, lon in enumerate(lons):
                lon = Observation.format_lon(lon)
                lat = Observation.format_lat(lats[point])
                stid = str(stids[point])

                pos = lon + ":" + lat
                self.index_pos.update({pos: point})
                if stid != "NA":
                    self.index_stid.update({stid: point})

            times = times.tolist()
            lons = lons.tolist()
            lats = lats.tolist()
            stids = stids.tolist()
            elevs = elevs.tolist()
            values = values.tolist()
            varnames = varnames.tolist()

            return times, lons, lats, stids, elevs, values, varnames
        return [], [], [], [], [], [], []

    def matching_obs(self, my_obs):
        """Match the observations.

        Args:
            my_obs (Observation): Observation to match.

        Returns:
            bool: True if found

        """
        found = False
        for i in range(0, len(self.observations)):
            if my_obs.obstime == self.observations.obstimes[i]:
                lon = self.observations.obstimes[i]
                lat = self.observations.obstimes[i]
                pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)

                if pos in self.index_pos:
                    found = True
        return found

    def points(self, geo, validtime=None):
        """Extract points from observations.

        Args:
            geo (surfex.Geo): Surfex geometry
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Returns:
            np.ndarray: Values in points.
        """
        my_times = []
        my_values = []
        my_stids = []
        times, lons, lats, stids, __, values, __ = self.get_obs()

        for i in range(0, geo.nlons):
            lon = geo.lonlist[i]
            lat = geo.latlist[i]
            pos = Observation.format_lon(lon) + ":" + Observation.format_lat(lat)

            lons.append(lon)
            lats.append(lat)
            if pos in self.index_pos:
                ind = self.index_pos[pos]
                if validtime is not None:
                    print("No time check implemented yet")
                my_times.append(times[ind])
                my_stids.append(stids[ind])
                my_values.append(values[ind])
            else:
                my_times.append(None)
                my_stids.append("NA")
                my_values.append(np.nan)
                logging.info("Could not find position %s in this data source", pos)

        my_values = np.asanyarray(my_values)
        return my_times, my_values, my_stids

    def write_json_file(self, filename, indent=None):
        """Write a json file.

        Args:
            filename (str): Name of file
            indent (int, optional): Indentation in file. Defaults to None.
        """
        obs2vectors = np.vectorize(Observation.obs2vectors)
        data = {}
        if len(self.observations) > 0:
            obstimes, lons, lats, stids, elevs, values, varnames = obs2vectors(
                self.observations
            )
            for obs, lon in enumerate(lons):
                data.update(
                    {
                        obs: {
                            "obstime": obstimes[obs].strftime("%Y%m%d%H%M%S"),
                            "varname": varnames[obs],
                            "lon": lon,
                            "lat": lats[obs],
                            "stid": stids[obs],
                            "elev": elevs[obs],
                            "value": values[obs],
                        }
                    }
                )
        with open(filename, mode="w", encoding="utf-8") as file_handler:
            json.dump(data, file_handler, indent=indent)


class NetatmoObservationSet(ObservationSet):
    """Observation set from netatmo."""

    def __init__(
        self,
        filenames,
        variable,
        target_time,
        dt=3600,
        re=True,
        lonrange=None,
        latrange=None,
        label="netatmo",
    ):
        """Construct netatmo obs.

        Args:
            filenames (list): Filenames
            variable (str): Variable
            target_time (as_datetime): _description_
            dt (int, optional): _description_. Defaults to 3600.
            re (bool, optional): _description_. Defaults to True.
            lonrange (_type_, optional): _description_. Defaults to None.
            latrange (_type_, optional): _description_. Defaults to None.
            label (str, optional): _description_. Defaults to "netatmo".

        Raises:
            RuntimeError: Lonrange must be a list with length 2
            RuntimeError: Latrange must be a list with length 2

        """
        if lonrange is None:
            lonrange = [-180, 180]

        if not isinstance(lonrange, list) or len(lonrange) != 2:
            raise RuntimeError(f"Lonrange must be a list with length 2 {lonrange}")

        if latrange is None:
            latrange = [-90, 90]

        if not isinstance(latrange, list) or len(latrange) != 2:
            raise RuntimeError(f"Latrange must be a list with length 2 {latrange}")

        data = {}  # key: id, value: list of values
        times = {}  # key: id, value: list of times
        metadata = {}

        # Parse json data
        #
        # The raw data is not valid JSON, since it is missing commas between lists
        # e.g. [...][...][...]. Instead format it like this: [..., ..., ...]

        observations = []
        num_missing_metadata = 0
        num_missing_obs = 0
        num_missing_time = 0
        num_missing_elev = 0
        num_wrong_time = 0
        for ifilename in filenames:
            with open(ifilename, mode="r", encoding="utf-8") as ifile:
                text = ifile.read()

            try:
                # Older files had an additional bug. This code will work for both cases
                if len(text) == 0:
                    # Sometimes netatmo files are empty
                    logging.info("Empty file: %s", ifilename)
                    continue

                if text[0] == "{":
                    text = f"[{text}"

                text = text.replace("}]{", "}{")
                text = text.replace("}][{", "},{")
                text = text.replace("}{", "},{")
                print(text)
                text = '{"data": %s}' % text
                raw = json.loads(text)
                raw = raw["data"]
                logging.debug("Parsing %d stations in %s", len(raw), ifilename)
            except RuntimeError:
                logging.error("Could not parse %s.", ifilename)
                continue

            for line in raw:
                if "data" in line and "_id" in line and "location" in line:
                    my_id = line["_id"]
                    location = line["location"]
                    curr_data = line["data"]
                    if variable in curr_data:
                        if "time_utc" in curr_data:
                            time_utc = curr_data["time_utc"]

                            # Record this observation if it is closer to the target time than any
                            # other previously parsed observation for this location. Also, only
                            # ecord if the time difference is within acceptable limits.

                            if "altitude" not in line:
                                num_missing_elev += 1

                            if not re or "altitude" in line:
                                lon = location[0]
                                lat = location[1]
                                if lonrange[0] <= lon <= lonrange[1]:
                                    if latrange[0] <= lat <= latrange[1]:
                                        if my_id not in data:
                                            data[my_id] = list()
                                            times[my_id] = list()

                                            elev = np.nan
                                            metadata[my_id] = {
                                                "lon": lon,
                                                "lat": lat,
                                                "elev": elev,
                                            }
                                        if (
                                            np.isnan(metadata[my_id]["elev"])
                                            and "altitude" in line
                                        ):
                                            metadata[my_id]["elev"] = line["altitude"]

                                        value = curr_data[variable]
                                        if variable == "Temperature":
                                            value = value + 273.15
                                        if variable == "Humidity":
                                            value = value * 0.01
                                        data[my_id] += [value]
                                        times[my_id] += [time_utc]
                        else:
                            num_missing_time += 1
                    else:
                        num_missing_obs += 1
                else:
                    num_missing_metadata += 1

        if target_time is not None:
            num_valid_stations = 0
            for my_id, time in times.items():
                this_diff_times = [
                    (utcfromtimestamp(t) - target_time).total_seconds() for t in time
                ]
                curr_times = [utcfromtimestamp(t) for t in time]
                if np.min(np.abs(np.array(this_diff_times))) < dt:
                    ibest = int(np.argmin(np.abs(np.array(this_diff_times))))
                    curr_time = curr_times[ibest]
                    elev = metadata[my_id]["elev"]
                    observations.append(
                        Observation(
                            curr_time,
                            metadata[my_id]["lon"],
                            metadata[my_id]["lat"],
                            data[my_id][ibest],
                            elev=elev,
                            varname=variable,
                        )
                    )
                    num_valid_stations += 1
        else:
            num_valid_stations = len(data)

        logging.info("Found %d valid observations:", num_valid_stations)
        logging.info("   %d missing obs", num_missing_obs)
        logging.info("   %d missing metadata", num_missing_metadata)
        logging.info("   %d missing timestamp", num_missing_time)
        logging.info("   %d wrong timestamp", num_wrong_time)
        if not re:
            extra = " (not removed)"
        else:
            extra = ""
        logging.debug("   %d missing elev%s", num_missing_elev, extra)

        ObservationSet.__init__(self, observations, label=label)


class MetFrostObservations(ObservationSet):
    """Observations from MET-Norway obs API (frost)."""

    def __init__(
        self,
        varname,
        stations=None,
        level=None,
        num_tries=3,
        wmo=None,
        providers=None,
        xproviders=None,
        blacklist=None,
        validtime=None,
        dt=3600,
        lonrange=None,
        latrange=None,
        unit=None,
        label="frost",
    ):
        """Construct obs set from Frost.

        Args:
            varname (_type_): _description_
            stations (_type_, optional): _description_. Defaults to None.
            level (_type_, optional): _description_. Defaults to None.
            num_tries (int, optional): _description_. Defaults to 3.
            wmo (_type_, optional): _description_. Defaults to None.
            providers (_type_, optional): _description_. Defaults to None.
            xproviders (_type_, optional): _description_. Defaults to None.
            blacklist (_type_, optional): _description_. Defaults to None.
            validtime (_type_, optional): _description_. Defaults to None.
            dt (int, optional): _description_. Defaults to 3600.
            lonrange (_type_, optional): _description_. Defaults to None.
            latrange (_type_, optional): _description_. Defaults to None.
            unit (_type_, optional): _description_. Defaults to None.
            label (str, optional): _description_. Defaults to "frost".

        Raises:
            KeyError: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_
            Exception: _description_

        """
        if blacklist is None:
            blacklist = []

        # extract client ID from environment variable
        if "CLIENTID" not in os.environ:
            raise KeyError("error: CLIENTID not found in environment\n")

        client_id = os.environ["CLIENTID"]

        # Get all the stations (in an area or country)
        # Make list of station IDs and dictionary of their lat,long,elev
        #
        ids = []
        tries = 1
        station_dict = dict()
        while tries <= num_tries:
            tries += 1
            # first get all the stations within a polygon or in a country?
            # 'geometry': 'POLYGON((10 60, 10 59, 11 60, 11 59))' # area around Oslo
            # 'country': 'NO' # everything in Norway

            parameters = {
                "types": "SensorSystem",
                "fields": "id,geometry,masl,wmoid,stationholders",
            }
            if lonrange is not None and latrange is not None:
                parameters.update(
                    {
                        "geometry": "POLYGON(("
                        + str(lonrange[0])
                        + " "
                        + str(latrange[0])
                        + ", "
                        + str(lonrange[0])
                        + " "
                        + str(latrange[1])
                        + ", "
                        + str(lonrange[1])
                        + " "
                        + str(latrange[1])
                        + ", "
                        + str(lonrange[1])
                        + " "
                        + str(latrange[0])
                        + " ))"
                    }
                )
            logging.debug("Request parameters: %s", str(parameters))
            req = requests.get(
                "https://frost.met.no/sources/v0.jsonld",
                parameters,
                auth=(client_id, ""),
                timeout=30,
            )

            logging.debug(
                "Request https://frost.met.no/sources/v0.jsonld returned %s",
                req.status_code,
            )
            # extract list of stations (if response was valid)
            if req.status_code == 200:
                data = req.json()["data"]
                ids = []
                count_discard = 0
                print(data)
                for data_block in data:
                    my_id = data_block["id"]
                    if "masl" in data_block:
                        elev = data_block["masl"]
                    else:
                        elev = -999  # missing value

                    # filter data for WMO and non WMO
                    keep_this_id = True
                    if "wmoId" in data_block and wmo is not None and wmo == 0:
                        # station is WMO skip
                        keep_this_id = False
                        logging.debug("throwing out this id (is WMO): %s", my_id)
                    elif "wmoId" not in data_block and wmo is not None and wmo == 1:
                        # station is not WMO skip
                        keep_this_id = False
                        logging.debug("throwing out this id (not WMO): %s", my_id)

                    # filter out stations with incomplete data
                    if keep_this_id and "geometry" not in data_block:
                        keep_this_id = False
                        logging.debug("throwing out this id (no geometry): %s", my_id)

                    # filters for station holders
                    if "stationHolders" not in data_block:
                        keep_this_id = False
                        logging.debug(
                            "throwing out this id (no stationHolders): %s", my_id
                        )
                    # select station providers
                    elif providers is not None:
                        providers = providers.split(",")
                        station_holders = data_block["stationHolders"]
                        if not any(x in station_holders for x in providers):
                            keep_this_id = False
                            logging.debug(
                                "throwing out this id (station holder): %s",
                                str(station_holders),
                            )
                    # or exclude certain station providers
                    elif xproviders is not None:
                        xproviders = xproviders.split(",")
                        station_holders = data_block["stationHolders"]
                        if any(x in station_holders for x in xproviders):
                            keep_this_id = False
                            logging.debug(
                                "throwing out this id (exclude station holder): %s",
                                str(station_holders),
                            )

                    # filter out blacklisted stations
                    if my_id in blacklist:
                        keep_this_id = False
                        logging.debug("throwing out blacklisted id: %s", my_id)

                    if stations is not None:
                        if my_id not in stations:
                            keep_this_id = False
                            logging.debug(
                                "Throwing out station because not in station list %s",
                                my_id,
                            )

                    logging.debug(
                        "Keep this ID: %s bool: %s", str(my_id), str(keep_this_id)
                    )
                    if keep_this_id:  # write into dict
                        ids.append(my_id)
                        # create a dictionary for these stations to store lat,long,elev for each
                        station_dict[my_id] = [
                            data_block["geometry"]["coordinates"][1],
                            data_block["geometry"]["coordinates"][0],
                            elev,
                        ]
                    else:
                        count_discard = count_discard + 1
                logging.debug("Number of stations: %s", str(len(ids)))
                logging.debug(
                    "Number of stations , debug=debugdiscarded: %s", str(count_discard)
                )
                break

            if req.status_code == 404:
                print("STATUS: No data was found for the list of query Ids.")
                break
            if tries > num_tries:
                raise Exception("ERROR: could not retrieve observations.")

        #
        # Use the station ID list to get the observation for each station
        #
        ids_obs_dict = {}  # declare outside loop, since may be more than one request
        # check how long the list of stations is and potentially break it up to shorten
        observations = []
        it_ids = len(ids)
        dt = as_timedelta(seconds=dt)
        while it_ids > 0:
            if it_ids > 50:
                # get last 50
                sub_id_list = ids[it_ids - 50 : it_ids]
                it_ids = it_ids - 50
            else:
                # get the rest if <50
                sub_id_list = ids[:it_ids]
                it_ids = 0

            tries = 1
            while tries <= num_tries:
                tries += 1
                # use the list of stations and get the observations for those
                parameters2 = {"sources": ",".join(sub_id_list), "elements": varname}
                date = validtime.strftime("%Y%m%d")
                hour = validtime.strftime("%H")

                # if have specified a date and time
                if date is not None and hour is not None:
                    # make these into a format that works for FROST
                    date_string = date
                    hour_string = hour
                    date_string_frost = (
                        date_string[0:4]
                        + "-"
                        + date_string[4:6]
                        + "-"
                        + date_string[6:8]
                        + "T"
                        + hour_string
                    )
                    parameters2["referencetime"] = date_string_frost
                # do not have date and time, so use latest
                else:
                    parameters2["referencetime"] = "latest"
                    parameters2["maxage"] = "PT30M"
                    parameters2["limit"] = 1

                logging.debug("Request parameters2: %s", str(parameters2))
                req = requests.get(
                    "https://frost.met.no/observations/v0.jsonld",
                    parameters2,
                    auth=(client_id, ""),
                    timeout=30,
                )

                logging.debug(
                    "Request https://frost.met.no/observations/v0.jsonld returned %s",
                    req.status_code,
                )
                if req.status_code == 200:
                    data = req.json()["data"]
                    for data_block in data:
                        # Check that reference time is ok, since sometimes future observations
                        # can be present when 'latest' is chosen for reference time
                        ref_str = data_block["referenceTime"]
                        ref_year = int(ref_str[0:4])
                        ref_month = int(ref_str[5:7])
                        ref_day = int(ref_str[8:10])
                        ref_hour = int(ref_str[11:13])
                        ref_min = int(ref_str[14:16])
                        ref_sec = int(ref_str[17:19])
                        ref_time = as_datetime_args(
                            year=ref_year,
                            month=ref_month,
                            day=ref_day,
                            hour=ref_hour,
                            minute=ref_min,
                            second=ref_sec,
                        )
                        logging.debug("ref_time %s validtime %s", ref_time, validtime)

                        read_unit = None
                        levels_ok = True
                        if "observations" in data_block:
                            for obs in data_block["observations"]:
                                logging.debug("%s", obs)
                                if "unit" in obs:
                                    read_unit = obs["unit"]
                                if "level" in obs:
                                    if level is not None:
                                        logging.debug("level %s", obs["level"])
                                        all_found = True
                                        for key in level:
                                            if key in obs["level"]:
                                                if str(level[key]) != str(
                                                    obs["level"][key]
                                                ):
                                                    logging.debug(
                                                        "%s != %s",
                                                        level[key],
                                                        obs["level"][key],
                                                    )
                                                    all_found = False
                                                else:
                                                    logging.debug(
                                                        "%s == %s",
                                                        level[key],
                                                        obs["level"][key],
                                                    )
                                        if not all_found:
                                            levels_ok = False

                        keep_this_obs = False
                        if levels_ok:
                            if abs(ref_time - validtime) < dt:
                                keep_this_obs = True

                        if keep_this_obs:
                            logging.debug("Keep this obs")
                            value = data_block["observations"][0]["value"]
                            if len(str(value)) > 0:  # not all stations have observations
                                source_id = str(data_block["sourceId"])
                                my_id = source_id.split(":")
                                if unit is not None:
                                    if read_unit is not None:
                                        if cfunits is None:
                                            raise Exception("cfunits not loaded!")
                                        read_unit = cfunits.Units(read_unit)
                                        unit = cfunits.Units(unit)
                                        value = cfunits.Units.conform(
                                            value, read_unit, unit
                                        )
                                    else:
                                        raise Exception("Did not read a unit to convert!")
                                ids_obs_dict[my_id[0]] = value

                    logging.debug(
                        "Station list length: %s, total number of observations "
                        "retrieved: %s",
                        str(len(sub_id_list)),
                        str(len(ids_obs_dict)),
                    )
                    break
                if req.status_code == 404:
                    print("STATUS: No data was found for the list of query Ids.")
                    break
                if tries > num_tries:
                    raise Exception("ERROR: could not retrieve observations.")

            for station, station_id in ids_obs_dict.items():
                value = float(station_id)
                id_info = station_dict[station]
                stid = str(station)[2:]
                lat = id_info[0]
                lon = id_info[1]
                elev = id_info[2]
                observations.append(
                    Observation(
                        validtime,
                        lon,
                        lat,
                        value,
                        stid=str(stid),
                        elev=elev,
                        varname=varname,
                    )
                )

        ObservationSet.__init__(self, observations, label=label)


class JsonObservationSet(ObservationSet):
    """JSON observation set."""

    def __init__(self, filename, label="json", var=None):
        """Construct an observation data set from a json file.

        Args:
            filename (str): Filename
            label (str, optional): Label of set. Defaults to "json".
            var (str, optional): Variable name. Defaults to None.

        Raises:
            RuntimeError: Varname is not found

        """
        with open(filename, mode="r", encoding="utf-8") as file_handler:
            obs = json.load(file_handler)
        observations = []
        for i in range(0, len(obs)):
            ind = str(i)
            obstime = as_datetime(obs[ind]["obstime"])
            lon = obs[ind]["lon"]
            lat = obs[ind]["lat"]
            elev = obs[ind]["elev"]
            value = obs[ind]["value"]
            stid = obs[ind]["stid"]
            varname = ""
            if "varname" in obs[ind]:
                varname = obs[ind]["varname"]

            if varname == "" and var is not None:
                raise RuntimeError("Varname is not found " + varname)

            if var is None or var == varname:
                observations.append(
                    Observation(
                        obstime, lon, lat, value, stid=stid, elev=elev, varname=varname
                    )
                )

        ObservationSet.__init__(self, observations, label=label)


class ObservationFromTitanJsonFile(ObservationSet):
    """Observation set from titan json file."""

    def __init__(self, an_time, filename, label=""):
        """Constuct obs set from a titan json file.

        Args:
            an_time (_type_): _description_
            filename (_type_): _description_
            label (str, optional): _description_. Defaults to "".

        """
        qc_obs = dataset_from_file(an_time, filename)
        observations = []
        for i in range(0, len(qc_obs)):
            observations.append(
                Observation(
                    qc_obs.obstimes[i],
                    qc_obs.lons[i],
                    qc_obs.lats[i],
                    qc_obs.elevs[i],
                    qc_obs.values[i],
                )
            )

        ObservationSet.__init__(self, observations, label=label)


def snow_pseudo_obs_cryoclim(
    validtime,
    grid_snow_class,
    grid_lons,
    grid_lats,
    step,
    fg_geo,
    grid_snow_fg,
    fg_threshold=2.0,
    new_snow_depth=0.01,
):
    """Cryoclim snow.

    Args:
        validtime (_type_): _description_
        grid_snow_class (_type_): _description_
        grid_lons (_type_): _description_
        grid_lats (_type_): _description_
        step (_type_): _description_
        fg_geo (_type_): _description_
        grid_snow_fg (_type_): _description_
        fg_threshold (float, optional): _description_. Defaults to 2.0.
        new_snow_depth (float, optional): _description_. Defaults to 0.01.

    Returns:
        _type_: _description_
    """
    n_x = grid_lons.shape[0]
    n_y = grid_lons.shape[1]

    n_x = int(n_x / step)
    n_y = int(n_y / step)

    # TODO rewrite to use lonlatvals geo
    counter = 0
    iii = 0
    res_lons = []
    res_lats = []
    p_snow_class = {}
    for __ in range(0, n_x):
        jjj = 0
        for __ in range(0, n_y):
            res_lons.append(grid_lons[iii, jjj])
            res_lats.append(grid_lats[iii, jjj])
            p_snow_class.update({str(counter): grid_snow_class[0, iii, jjj]})
            counter = counter + 1
            jjj = jjj + step
        iii = iii + step

    p_fg_snow_depth = gridpos2points(
        fg_geo.lons, fg_geo.lats, np.asarray(res_lons), np.asarray(res_lats), grid_snow_fg
    )
    in_grid = inside_grid(
        fg_geo.lons,
        fg_geo.lats,
        np.asarray(res_lons),
        np.asarray(res_lats),
        distance=2500.0,
    )

    # Ordering of points must be the same.....
    obs = []
    flags = []
    cis = []
    lafs = []
    providers = []
    for i in range(0, p_fg_snow_depth.shape[0]):

        p_snow_fg = p_fg_snow_depth[i]
        logging.debug("%s %s %s %s", i, p_snow_fg, res_lons[i], res_lats[i])
        if not np.isnan(p_snow_fg):
            # Check if in grid
            if in_grid[i]:
                obs_value = np.nan
                if p_snow_class[str(i)] == 1:
                    if p_snow_fg > 0:
                        if fg_threshold is not None:
                            if p_snow_fg <= fg_threshold:
                                obs_value = p_snow_fg
                        else:
                            obs_value = p_snow_fg
                    else:
                        obs_value = new_snow_depth
                elif p_snow_class[str(i)] == 0:
                    if p_snow_fg > 0:
                        obs_value = 0.0

                if not np.isnan(obs_value):
                    flags.append(0)
                    cis.append(0)
                    lafs.append(0)
                    providers.append(0)
                    obs.append(
                        Observation(validtime, res_lons[i], res_lats[i], obs_value)
                    )

    logging.info("Possible pseudo-observations: %s", n_x * n_y)
    logging.info("Pseudo-observations created: %s", len(obs))
    return QCDataSet(validtime, obs, flags, cis, lafs, providers)


def sm_obs_sentinel(
    validtime,
    grid_sm_class,
    grid_lons,
    grid_lats,
    step,
    fg_geo,
    grid_sm_fg,
    fg_threshold=1.0,
):
    """Sentinel.

    Args:
        validtime (_type_): _description_
        grid_sm_class (_type_): _description_
        grid_lons (_type_): _description_
        grid_lats (_type_): _description_
        step (_type_): _description_
        fg_geo (_type_): _description_
        grid_sm_fg (_type_): _description_
        fg_threshold (_type_, optional): _description_. Defaults to 1..

    Returns:
        _type_: _description_

    """
    n_x = grid_lons.shape[0]
    n_y = grid_lons.shape[1]

    n_x = int(n_x / step)
    n_y = int(n_y / step)

    # TODO rewrite to use lonlatvals geo
    counter = 0
    iii = 0
    res_lons = []
    res_lats = []
    p_sm_class = {}
    for __ in range(0, n_x):
        jjj = 0
        for __ in range(0, n_y):
            res_lons.append(grid_lons[iii, jjj])
            res_lats.append(grid_lats[iii, jjj])
            p_sm_class.update({str(counter): grid_sm_class[iii, jjj]})
            counter = counter + 1
            jjj = jjj + step
        iii = iii + step

    p_fg_sm = gridpos2points(
        fg_geo.lons, fg_geo.lats, np.asarray(res_lons), np.asarray(res_lats), grid_sm_fg
    )
    in_grid = inside_grid(
        fg_geo.lons,
        fg_geo.lats,
        np.asarray(res_lons),
        np.asarray(res_lats),
        distance=2500.0,
    )
    # Ordering of points must be the same.....
    obs = []
    flags = []
    cis = []
    lafs = []
    providers = []
    for i in range(0, p_fg_sm.shape[0]):

        p_sm_fg = p_fg_sm[i]
        if not np.isnan(p_sm_fg):
            # Check if in grid
            if in_grid[i]:
                obs_value = np.nan
                if (p_sm_class[str(i)] > 1) or (p_sm_class[str(i)] < 0):
                    if p_sm_fg <= fg_threshold:
                        obs_value = p_sm_fg
                    else:
                        obs_value = 999

                else:
                    obs_value = p_sm_class[str(i)]

                if not np.isnan(obs_value):
                    flags.append(0)
                    cis.append(0)
                    lafs.append(0)
                    providers.append(0)
                    obs.append(
                        Observation(
                            validtime,
                            res_lons[i],
                            res_lats[i],
                            obs_value,
                            varname="surface_soil_moisture",
                        )
                    )

    logging.info("Possible pseudo-observations: %s", n_x * n_y)
    logging.info("Pseudo-observations created: %s", len(obs))
    return QCDataSet(validtime, obs, flags, cis, lafs, providers)
